from hummingbot.core.api_throttler.data_types import RateLimit

BASE_URL = "https://api.coingecko.com/api/v3"
PING_REST_ENDPOINT = "/ping"
PRICES_REST_ENDPOINT = "/coins/markets"
SUPPORTED_VS_TOKENS_REST_ENDPOINT = "/simple/supported_vs_currencies"

COOLOFF_AFTER_BAN = 60.0 * 1.05

REST_CALL_RATE_LIMIT_ID = "coin_gecko_rest_rate_limit_id"
RATE_LIMITS = [RateLimit(REST_CALL_RATE_LIMIT_ID, limit=10, time_interval=60)]

TOKEN_CATEGORIES = [
    "cryptocurrency",
    "decentralized-exchange",
    "decentralized-finance-defi",
    "smart-contract-platform",
    "stablecoins",
    "wrapped-tokens",
]



import asyncio
import hashlib
import time
from decimal import Decimal
from typing import TYPE_CHECKING, Any, AsyncIterable, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.derivative.hyperliquid_perpetual import (
    hyperliquid_perpetual_constants as CONSTANTS,
    hyperliquid_perpetual_web_utils as web_utils,
)
from hummingbot.connector.derivative.hyperliquid_perpetual.hyperliquid_perpetual_api_order_book_data_source import (
    HyperliquidPerpetualAPIOrderBookDataSource,
)
from hummingbot.connector.derivative.hyperliquid_perpetual.hyperliquid_perpetual_auth import HyperliquidPerpetualAuth
from hummingbot.connector.derivative.hyperliquid_perpetual.hyperliquid_perpetual_user_stream_data_source import (
    HyperliquidPerpetualUserStreamDataSource,
)
from hummingbot.connector.derivative.position import Position
from hummingbot.connector.perpetual_derivative_py_base import PerpetualDerivativePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair, get_new_client_order_id
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, PositionSide, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future, safe_gather
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter

bpm_logger = None


class HyperliquidPerpetualDerivative(PerpetualDerivativePyBase):
    web_utils = web_utils

    SHORT_POLL_INTERVAL = 5.0
    LONG_POLL_INTERVAL = 12.0

    def __init__(
            self,
            client_config_map: "ClientConfigAdapter",
            hyperliquid_perpetual_api_secret: str = None,
            use_vault: bool = False,
            hyperliquid_perpetual_api_key: str = None,
            trading_pairs: Optional[List[str]] = None,
            trading_required: bool = True,
            domain: str = CONSTANTS.DOMAIN,
    ):
        self.hyperliquid_perpetual_api_key = hyperliquid_perpetual_api_key
        self.hyperliquid_perpetual_secret_key = hyperliquid_perpetual_api_secret
        self._use_vault = use_vault
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._domain = domain
        self._position_mode = None
        self._last_trade_history_timestamp = None
        self.coin_to_asset: Dict[str, int] = {}
        super().__init__(client_config_map)

    @property
    def name(self) -> str:
        # Note: domain here refers to the entire exchange name. i.e. hyperliquid_perpetual or hyperliquid_perpetual_testnet
        return self._domain

    @property
    def authenticator(self) -> Optional[HyperliquidPerpetualAuth]:
        if self._trading_required:
            return HyperliquidPerpetualAuth(self.hyperliquid_perpetual_api_key, self.hyperliquid_perpetual_secret_key,
                                            self._use_vault)
        return None

    @property
    def rate_limits_rules(self) -> List[RateLimit]:
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def client_order_id_max_length(self) -> int:
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self) -> str:
        return CONSTANTS.BROKER_ID

    @property
    def trading_rules_request_path(self) -> str:
        return CONSTANTS.EXCHANGE_INFO_URL

    @property
    def trading_pairs_request_path(self) -> str:
        return CONSTANTS.EXCHANGE_INFO_URL

    @property
    def check_network_request_path(self) -> str:
        return CONSTANTS.PING_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @property
    def funding_fee_poll_interval(self) -> int:
        return 120

    async def _make_network_check_request(self):
        await self._api_post(path_url=self.check_network_request_path, data={"type": CONSTANTS.META_INFO})

    def supported_order_types(self) -> List[OrderType]:
        """
        :return a list of OrderType supported by this connector
        """
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    def supported_position_modes(self):
        """
        This method needs to be overridden to provide the accurate information depending on the exchange.
        """
        return [PositionMode.ONEWAY]

    def get_buy_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.buy_order_collateral_token

    def get_sell_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.sell_order_collateral_token

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        return False

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            auth=self._auth)

    async def _make_trading_rules_request(self) -> Any:
        exchange_info = await self._api_post(path_url=self.trading_rules_request_path,
                                             data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})
        return exchange_info

    async def _make_trading_pairs_request(self) -> Any:
        exchange_info = await self._api_post(path_url=self.trading_pairs_request_path,
                                             data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})
        return exchange_info

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return CONSTANTS.ORDER_NOT_EXIST_MESSAGE in str(status_update_exception)

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return CONSTANTS.UNKNOWN_ORDER_MESSAGE in str(cancelation_exception)

    def quantize_order_price(self, trading_pair: str, price: Decimal) -> Decimal:
        """
        Applies trading rule to quantize order price.
        """
        d_price = Decimal(round(float(f"{price:.5g}"), 6))
        return d_price

    async def _update_trading_rules(self):
        exchange_info = await self._api_post(path_url=self.trading_rules_request_path,
                                             data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})
        trading_rules_list = await self._format_trading_rules(exchange_info)
        self._trading_rules.clear()
        for trading_rule in trading_rules_list:
            self._trading_rules[trading_rule.trading_pair] = trading_rule
        self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)

    async def _initialize_trading_pair_symbol_map(self):
        try:
            exchange_info = await self._api_post(path_url=self.trading_pairs_request_path,
                                                 data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})

            self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)
        except Exception:
            self.logger().exception("There was an error requesting exchange info.")

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return HyperliquidPerpetualAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return HyperliquidPerpetualUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    async def _status_polling_loop_fetch_updates(self):
        await safe_gather(
            self._update_trade_history(),
            self._update_order_status(),
            self._update_balances(),
            self._update_positions(),
        )

    async def _update_order_status(self):
        await self._update_orders()

    async def _update_lost_orders_status(self):
        await self._update_lost_orders()

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 position_action: PositionAction,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = is_maker or False
        fee = build_trade_fee(
            self.name,
            is_maker,
            base_currency=base_currency,
            quote_currency=quote_currency,
            order_type=order_type,
            order_side=order_side,
            amount=amount,
            price=price,
        )
        return fee

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=tracked_order.trading_pair)
        coin = symbol.split("-")[0]

        api_params = {
            "type": "cancel",
            "cancels": {
                "asset": self.coin_to_asset[coin],
                "cloid": order_id
            },
        }
        cancel_result = await self._api_post(
            path_url=CONSTANTS.CANCEL_ORDER_URL,
            data=api_params,
            is_auth_required=True)

        if cancel_result.get("status") == "err" or "error" in cancel_result["response"]["data"]["statuses"][0]:
            self.logger().debug(f"The order {order_id} does not exist on Hyperliquid Perpetuals. "
                                f"No cancelation needed.")
            await self._order_tracker.process_order_not_found(order_id)
            raise IOError(f'{cancel_result["response"]["data"]["statuses"][0]["error"]}')
        if "success" in cancel_result["response"]["data"]["statuses"][0]:
            return True
        return False

    # === Orders placing ===

    def buy(self,
            trading_pair: str,
            amount: Decimal,
            order_type=OrderType.LIMIT,
            price: Decimal = s_decimal_NaN,
            **kwargs) -> str:
        """
        Creates a promise to create a buy order using the parameters

        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price

        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = get_new_client_order_id(
            is_buy=True,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length
        )
        md5 = hashlib.md5()
        md5.update(order_id.encode('utf-8'))
        hex_order_id = f"0x{md5.hexdigest()}"
        if order_type is OrderType.MARKET:
            mid_price = self.get_mid_price(trading_pair)
            slippage = CONSTANTS.MARKET_ORDER_SLIPPAGE
            market_price = mid_price * Decimal(1 + slippage)
            price = self.quantize_order_price(trading_pair, market_price)

        safe_ensure_future(self._create_order(
            trade_type=TradeType.BUY,
            order_id=hex_order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return hex_order_id

    def sell(self,
             trading_pair: str,
             amount: Decimal,
             order_type: OrderType = OrderType.LIMIT,
             price: Decimal = s_decimal_NaN,
             **kwargs) -> str:
        """
        Creates a promise to create a sell order using the parameters.
        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price
        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = get_new_client_order_id(
            is_buy=False,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length
        )
        md5 = hashlib.md5()
        md5.update(order_id.encode('utf-8'))
        hex_order_id = f"0x{md5.hexdigest()}"
        if order_type is OrderType.MARKET:
            mid_price = self.get_mid_price(trading_pair)
            slippage = CONSTANTS.MARKET_ORDER_SLIPPAGE
            market_price = mid_price * Decimal(1 - slippage)
            price = self.quantize_order_price(trading_pair, market_price)

        safe_ensure_future(self._create_order(
            trade_type=TradeType.SELL,
            order_id=hex_order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return hex_order_id

    async def _place_order(
            self,
            order_id: str,
            trading_pair: str,
            amount: Decimal,
            trade_type: TradeType,
            order_type: OrderType,
            price: Decimal,
            position_action: PositionAction = PositionAction.NIL,
            **kwargs,
    ) -> Tuple[str, float]:

        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        coin = symbol.split("-")[0]
        param_order_type = {"limit": {"tif": "Gtc"}}
        if order_type is OrderType.LIMIT_MAKER:
            param_order_type = {"limit": {"tif": "Alo"}}
        if order_type is OrderType.MARKET:
            param_order_type = {"limit": {"tif": "Ioc"}}

        api_params = {
            "type": "order",
            "grouping": "na",
            "orders": {
                "asset": self.coin_to_asset[coin],
                "isBuy": True if trade_type is TradeType.BUY else False,
                "limitPx": float(price),
                "sz": float(amount),
                "reduceOnly": position_action == PositionAction.CLOSE,
                "orderType": param_order_type,
                "cloid": order_id,
            }
        }
        order_result = await self._api_post(
            path_url=CONSTANTS.CREATE_ORDER_URL,
            data=api_params,
            is_auth_required=True)
        if order_result.get("status") == "err":
            raise IOError(f"Error submitting order {order_id}: {order_result['response']}")
        else:
            o_order_result = order_result['response']["data"]["statuses"][0]
        if "error" in o_order_result:
            raise IOError(f"Error submitting order {order_id}: {o_order_result['error']}")
        o_data = o_order_result.get("resting") or o_order_result.get("filled")
        o_id = str(o_data["oid"])
        return (o_id, self.current_timestamp)

    async def _update_trade_history(self):
        orders = list(self._order_tracker.all_fillable_orders.values())
        all_fillable_orders = self._order_tracker.all_fillable_orders_by_exchange_order_id
        all_fills_response = []
        if len(orders) > 0:
            try:
                all_fills_response = await self._api_post(
                    path_url=CONSTANTS.ACCOUNT_TRADE_LIST_URL,
                    data={
                        "type": CONSTANTS.TRADES_TYPE,
                        "user": self.hyperliquid_perpetual_api_key,
                    })
            except asyncio.CancelledError:
                raise
            except Exception as request_error:
                self.logger().warning(
                    f"Failed to fetch trade updates. Error: {request_error}",
                    exc_info=request_error,
                )
            for trade_fill in all_fills_response:
                self._process_trade_rs_event_message(order_fill=trade_fill, all_fillable_order=all_fillable_orders)

    def _process_trade_rs_event_message(self, order_fill: Dict[str, Any], all_fillable_order):
        exchange_order_id = str(order_fill.get("oid"))
        fillable_order = all_fillable_order.get(exchange_order_id)
        if fillable_order is not None:
            fee_asset = fillable_order.quote_asset

            position_action = PositionAction.OPEN if order_fill["dir"].split(" ")[0] == "Open" else PositionAction.CLOSE
            fee = TradeFeeBase.new_perpetual_fee(
                fee_schema=self.trade_fee_schema(),
                position_action=position_action,
                percent_token=fee_asset,
                flat_fees=[TokenAmount(amount=Decimal(order_fill["fee"]), token=fee_asset)]
            )

            trade_update = TradeUpdate(
                trade_id=str(order_fill["tid"]),
                client_order_id=fillable_order.client_order_id,
                exchange_order_id=str(order_fill["oid"]),
                trading_pair=fillable_order.trading_pair,
                fee=fee,
                fill_base_amount=Decimal(order_fill["sz"]),
                fill_quote_amount=Decimal(order_fill["px"]) * Decimal(order_fill["sz"]),
                fill_price=Decimal(order_fill["px"]),
                fill_timestamp=order_fill["time"] * 1e-3,
            )

            self._order_tracker.process_trade_update(trade_update)

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        # Use _update_trade_history instead
        pass

    async def _handle_update_error_for_active_order(self, order: InFlightOrder, error: Exception):
        try:
            raise error
        except (asyncio.TimeoutError, KeyError):
            self.logger().debug(
                f"Tracked order {order.client_order_id} does not have an exchange id. "
                f"Attempting fetch in next polling interval."
            )
            await self._order_tracker.process_order_not_found(order.client_order_id)
        except asyncio.CancelledError:
            raise
        except Exception as request_error:
            self.logger().warning(
                f"Error fetching status update for the active order {order.client_order_id}: {request_error}.",
            )
            self.logger().debug(
                f"Order {order.client_order_id} not found counter: {self._order_tracker._order_not_found_records.get(order.client_order_id, 0)}")
            await self._order_tracker.process_order_not_found(order.client_order_id)

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        client_order_id = tracked_order.client_order_id
        try:
            if tracked_order.exchange_order_id:
                exchange_order_id = tracked_order.exchange_order_id
            else:
                exchange_order_id = await tracked_order.get_exchange_order_id()
        except asyncio.TimeoutError:
            exchange_order_id = None
        order_update = await self._api_post(
            path_url=CONSTANTS.ORDER_URL,
            data={
                "type": CONSTANTS.ORDER_STATUS_TYPE,
                "user": self.hyperliquid_perpetual_api_key,
                "oid": int(exchange_order_id) if exchange_order_id else client_order_id
            })
        current_state = order_update["order"]["status"]
        _exchange_order_id = str(tracked_order.exchange_order_id) if tracked_order.exchange_order_id else str(
            order_update["order"]["order"]["oid"])
        _order_update: OrderUpdate = OrderUpdate(
            trading_pair=tracked_order.trading_pair,
            update_timestamp=order_update["order"]["order"]["timestamp"] * 1e-3,
            new_state=CONSTANTS.ORDER_STATE[current_state],
            client_order_id=order_update["order"]["order"]["cloid"] or client_order_id,
            exchange_order_id=_exchange_order_id,
        )
        return _order_update

    async def _iter_user_event_queue(self) -> AsyncIterable[Dict[str, any]]:
        while True:
            try:
                yield await self._user_stream_tracker.user_stream.get()
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    "Unknown error. Retrying after 1 seconds.",
                    exc_info=True,
                    app_warning_msg="Could not fetch user events from Hyperliquid. Check API key and network connection.",
                )
                await self._sleep(1.0)

    async def _user_stream_event_listener(self):
        """
        Listens to messages from _user_stream_tracker.user_stream queue.
        Traders, Orders, and Balance updates from the WS.
        """
        user_channels = [
            CONSTANTS.USER_ORDERS_ENDPOINT_NAME,
            CONSTANTS.USEREVENT_ENDPOINT_NAME,
        ]
        async for event_message in self._iter_user_event_queue():
            try:
                if isinstance(event_message, dict):
                    channel: str = event_message.get("channel", None)
                    results = event_message.get("data", None)
                elif event_message is asyncio.CancelledError:
                    raise asyncio.CancelledError
                else:
                    raise Exception(event_message)
                if channel not in user_channels:
                    self.logger().error(
                        f"Unexpected message in user stream: {event_message}.", exc_info=True)
                    continue
                if channel == CONSTANTS.USER_ORDERS_ENDPOINT_NAME:
                    for order_msg in results:
                        self._process_order_message(order_msg)
                elif channel == CONSTANTS.USEREVENT_ENDPOINT_NAME:
                    if "fills" in results:
                        for trade_msg in results["fills"]:
                            await self._process_trade_message(trade_msg)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error(
                    "Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    async def _process_trade_message(self, trade: Dict[str, Any], client_order_id: Optional[str] = None):
        """
        Updates in-flight order and trigger order filled event for trade message received. Triggers order completed
        event if the total executed amount equals to the specified order amount.
        Example Trade:
        """
        exchange_order_id = str(trade.get("oid", ""))
        tracked_order = self._order_tracker.all_fillable_orders_by_exchange_order_id.get(exchange_order_id)

        if tracked_order is None:
            all_orders = self._order_tracker.all_fillable_orders
            for k, v in all_orders.items():
                await v.get_exchange_order_id()
            _cli_tracked_orders = [o for o in all_orders.values() if exchange_order_id == o.exchange_order_id]
            if not _cli_tracked_orders:
                self.logger().debug(f"Ignoring trade message with id {client_order_id}: not in in_flight_orders.")
                return
            tracked_order = _cli_tracked_orders[0]
        trading_pair_base_coin = tracked_order.base_asset
        if trade["coin"] == trading_pair_base_coin:
            position_action = PositionAction.OPEN if trade["dir"].split(" ")[0] == "Open" else PositionAction.CLOSE
            fee_asset = tracked_order.quote_asset
            fee = TradeFeeBase.new_perpetual_fee(
                fee_schema=self.trade_fee_schema(),
                position_action=position_action,
                percent_token=fee_asset,
                flat_fees=[TokenAmount(amount=Decimal(trade["fee"]), token=fee_asset)]
            )
            trade_update: TradeUpdate = TradeUpdate(
                trade_id=str(trade["tid"]),
                client_order_id=tracked_order.client_order_id,
                exchange_order_id=str(trade["oid"]),
                trading_pair=tracked_order.trading_pair,
                fill_timestamp=trade["time"] * 1e-3,
                fill_price=Decimal(trade["px"]),
                fill_base_amount=Decimal(trade["sz"]),
                fill_quote_amount=Decimal(trade["px"]) * Decimal(trade["sz"]),
                fee=fee,
            )
            self._order_tracker.process_trade_update(trade_update)

    def _process_order_message(self, order_msg: Dict[str, Any]):
        """
        Updates in-flight order and triggers cancelation or failure event if needed.

        :param order_msg: The order response from either REST or web socket API (they are of the same format)

        Example Order:
        """
        client_order_id = str(order_msg["order"].get("cloid", ""))
        tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)
        if not tracked_order:
            self.logger().debug(f"Ignoring order message with id {client_order_id}: not in in_flight_orders.")
            return
        current_state = order_msg["status"]
        tracked_order.update_exchange_order_id(str(order_msg["order"]["oid"]))
        order_update: OrderUpdate = OrderUpdate(
            trading_pair=tracked_order.trading_pair,
            update_timestamp=order_msg["statusTimestamp"] * 1e-3,
            new_state=CONSTANTS.ORDER_STATE[current_state],
            client_order_id=order_msg["order"]["cloid"],
            exchange_order_id=str(order_msg["order"]["oid"]),
        )
        self._order_tracker.process_order_update(order_update=order_update)

    async def _format_trading_rules(self, exchange_info_dict: List) -> List[TradingRule]:
        """
        Queries the necessary API endpoint and initialize the TradingRule object for each trading pair being traded.

        Parameters
        ----------
        exchange_info_dict:
            Trading rules dictionary response from the exchange
        """
        # rules: list = exchange_info_dict[0]
        self.coin_to_asset = {asset_info["name"]: asset for (asset, asset_info) in
                              enumerate(exchange_info_dict[0]["universe"])}

        coin_infos: list = exchange_info_dict[0]['universe']
        price_infos: list = exchange_info_dict[1]
        return_val: list = []
        for coin_info, price_info in zip(coin_infos, price_infos):
            try:
                ex_symbol = f'{coin_info["name"]}-{CONSTANTS.CURRENCY}'
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=ex_symbol)
                step_size = Decimal(str(10 ** -coin_info.get("szDecimals")))

                price_size = Decimal(str(10 ** -len(price_info.get("markPx").split('.')[1])))
                _min_order_size = Decimal(str(10 ** -len(price_info.get("openInterest").split('.')[1])))
                collateral_token = CONSTANTS.CURRENCY
                return_val.append(
                    TradingRule(
                        trading_pair,
                        min_base_amount_increment=step_size,
                        min_price_increment=price_size,
                        min_order_size=_min_order_size,
                        buy_order_collateral_token=collateral_token,
                        sell_order_collateral_token=collateral_token,
                    )
                )
            except Exception:
                self.logger().error(f"Error parsing the trading pair rule {exchange_info_dict}. Skipping.",
                                    exc_info=True)
        return return_val

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: List):
        mapping = bidict()
        for symbol_data in filter(web_utils.is_exchange_information_valid, exchange_info[0].get("universe", [])):
            exchange_symbol = f'{symbol_data["name"]}-{CONSTANTS.CURRENCY}'
            base = symbol_data["name"]
            quote = CONSTANTS.CURRENCY
            trading_pair = combine_to_hb_trading_pair(base, quote)
            if trading_pair in mapping.inverse:
                self._resolve_trading_pair_symbols_duplicate(mapping, exchange_symbol, base, quote)
            else:
                mapping[exchange_symbol] = trading_pair
        self._set_trading_pair_symbol_map(mapping)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        exchange_symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        coin = exchange_symbol.split("-")[0]
        response = await self._api_post(path_url=CONSTANTS.TICKER_PRICE_CHANGE_URL,
                                        data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})
        price = 0
        for index, i in enumerate(response[0]['universe']):
            if i['name'] == coin:
                price = float(response[1][index]['markPx'])
        return price

    def _resolve_trading_pair_symbols_duplicate(self, mapping: bidict, new_exchange_symbol: str, base: str, quote: str):
        """Resolves name conflicts provoked by futures contracts.

        If the expected BASEQUOTE combination matches one of the exchange symbols, it is the one taken, otherwise,
        the trading pair is removed from the map and an error is logged.
        """
        expected_exchange_symbol = f"{base}{quote}"
        trading_pair = combine_to_hb_trading_pair(base, quote)
        current_exchange_symbol = mapping.inverse[trading_pair]
        if current_exchange_symbol == expected_exchange_symbol:
            pass
        elif new_exchange_symbol == expected_exchange_symbol:
            mapping.pop(current_exchange_symbol)
            mapping[new_exchange_symbol] = trading_pair
        else:
            self.logger().error(
                f"Could not resolve the exchange symbols {new_exchange_symbol} and {current_exchange_symbol}")
            mapping.pop(current_exchange_symbol)

    async def _update_balances(self):
        """
        Calls the REST API to update total and available balances.
        """

        account_info = await self._api_post(path_url=CONSTANTS.ACCOUNT_INFO_URL,
                                            data={"type": CONSTANTS.USER_STATE_TYPE,
                                                  "user": self.hyperliquid_perpetual_api_key},
                                            )
        quote = CONSTANTS.CURRENCY
        self._account_balances[quote] = Decimal(account_info["crossMarginSummary"]["accountValue"])
        self._account_available_balances[quote] = Decimal(account_info["withdrawable"])

    async def _update_positions(self):
        positions = await self._api_post(path_url=CONSTANTS.POSITION_INFORMATION_URL,
                                         data={"type": CONSTANTS.USER_STATE_TYPE,
                                               "user": self.hyperliquid_perpetual_api_key}
                                         )
        for position in positions["assetPositions"]:
            position = position.get("position")
            ex_trading_pair = position.get("coin") + "-" + CONSTANTS.CURRENCY
            hb_trading_pair = await self.trading_pair_associated_to_exchange_symbol(ex_trading_pair)

            position_side = PositionSide.LONG if Decimal(position.get("szi")) > 0 else PositionSide.SHORT
            unrealized_pnl = Decimal(position.get("unrealizedPnl"))
            entry_price = Decimal(position.get("entryPx"))
            amount = Decimal(position.get("szi", 0))
            leverage = Decimal(position.get("leverage").get("value"))
            pos_key = self._perpetual_trading.position_key(hb_trading_pair, position_side)
            if amount != 0:
                _position = Position(
                    trading_pair=hb_trading_pair,
                    position_side=position_side,
                    unrealized_pnl=unrealized_pnl,
                    entry_price=entry_price,
                    amount=amount,
                    leverage=leverage
                )
                self._perpetual_trading.set_position(pos_key, _position)
            else:
                self._perpetual_trading.remove_position(pos_key)
        if not positions.get("assetPositions"):
            keys = list(self._perpetual_trading.account_positions.keys())
            for key in keys:
                self._perpetual_trading.remove_position(key)

    async def _get_position_mode(self) -> Optional[PositionMode]:
        return PositionMode.ONEWAY

    async def _trading_pair_position_mode_set(self, mode: PositionMode, trading_pair: str) -> Tuple[bool, str]:
        msg = ""
        success = True
        initial_mode = await self._get_position_mode()
        if initial_mode != mode:
            msg = "hyperliquid only supports the ONEWAY position mode."
            success = False
        return success, msg

    async def _set_trading_pair_leverage(self, trading_pair: str, leverage: int) -> Tuple[bool, str]:
        coin = trading_pair.split("-")[0]
        if not self.coin_to_asset:
            await self._update_trading_rules()
        params = {
            "type": "updateLeverage",
            "asset": self.coin_to_asset[coin],
            "isCross": True,
            "leverage": leverage,
        }
        try:
            set_leverage = await self._api_post(
                path_url=CONSTANTS.SET_LEVERAGE_URL,
                data=params,
                is_auth_required=True)
            success = False
            msg = ""
            if set_leverage.get("status") == "err":
                raise IOError(f"{set_leverage}")
            if set_leverage["status"] == 'ok':
                success = True
            else:
                msg = 'Unable to set leverage'
            return success, msg
        except Exception as exception:
            success = False
            msg = f"There was an error setting the leverage for {trading_pair} ({exception})"

        return success, msg

    async def _fetch_last_fee_payment(self, trading_pair: str) -> Tuple[int, Decimal, Decimal]:
        exchange_symbol = await self.exchange_symbol_associated_to_pair(trading_pair)
        coin = exchange_symbol.split("-")[0]

        funding_info_response = await self._api_post(path_url=CONSTANTS.GET_LAST_FUNDING_RATE_PATH_URL,
                                                     data={
                                                         "type": "userFunding",
                                                         "user": self.hyperliquid_perpetual_api_key,
                                                         "startTime": self._last_funding_time(),
                                                     }
                                                     )
        sorted_payment_response = [i for i in funding_info_response if i["delta"]["coin"] == coin]
        if len(sorted_payment_response) < 1:
            timestamp, funding_rate, payment = 0, Decimal("-1"), Decimal("-1")
            return timestamp, funding_rate, payment
        funding_payment = sorted_payment_response[0]
        _payment = Decimal(funding_payment["delta"]["usdc"])
        funding_rate = Decimal(funding_payment["delta"]["fundingRate"])
        timestamp = funding_payment["time"] * 1e-3
        if _payment != Decimal("0"):
            payment = _payment
        else:
            timestamp, funding_rate, payment = 0, Decimal("-1"), Decimal("-1")
        return timestamp, funding_rate, payment

    def _last_funding_time(self) -> int:
        """
        Funding settlement occurs every 1 hours as mentioned in https://hyperliquid.gitbook.io/hyperliquid-docs/trading/funding
        """
        return int(((time.time() // 3600) - 1) * 3600 * 1e3)



import pandas as pd

from hummingbot.core.data_type.common import TradeType
from hummingbot.strategy_v2.backtesting.executor_simulator_base import ExecutorSimulation, ExecutorSimulatorBase
from hummingbot.strategy_v2.executors.position_executor.data_types import PositionExecutorConfig
from hummingbot.strategy_v2.models.executors import CloseType


class PositionExecutorSimulator(ExecutorSimulatorBase):
    def simulate(self, df: pd.DataFrame, config: PositionExecutorConfig, trade_cost: float) -> ExecutorSimulation:
        if config.triple_barrier_config.open_order_type.is_limit_type():
            entry_condition = (df['close'] <= config.entry_price) if config.side == TradeType.BUY else (df['close'] >= config.entry_price)
            start_timestamp = df[entry_condition]['timestamp'].min()
        else:
            start_timestamp = df['timestamp'].min()
        last_timestamp = df['timestamp'].max()

        # Set up barriers
        tp = float(config.triple_barrier_config.take_profit) if config.triple_barrier_config.take_profit else None
        trailing_sl_trigger_pct = None
        trailing_sl_delta_pct = None
        if config.triple_barrier_config.trailing_stop:
            trailing_sl_trigger_pct = float(config.triple_barrier_config.trailing_stop.activation_price)
            trailing_sl_delta_pct = float(config.triple_barrier_config.trailing_stop.trailing_delta)
        tl = config.triple_barrier_config.time_limit if config.triple_barrier_config.time_limit else None
        tl_timestamp = config.timestamp + tl if tl else last_timestamp

        # Filter dataframe based on the conditions
        df_filtered = df[df['timestamp'] <= tl_timestamp].copy()
        df_filtered['net_pnl_pct'] = 0.0
        df_filtered['net_pnl_quote'] = 0.0
        df_filtered['cum_fees_quote'] = 0.0
        df_filtered['filled_amount_quote'] = 0.0
        df_filtered["current_position_average_price"] = float(config.entry_price)

        if pd.isna(start_timestamp):
            return ExecutorSimulation(config=config, executor_simulation=df_filtered, close_type=CloseType.TIME_LIMIT)

        entry_price = df.loc[df['timestamp'] == start_timestamp, 'close'].values[0]
        side_multiplier = 1 if config.side == TradeType.BUY else -1

        returns_df = df_filtered[df_filtered['timestamp'] >= start_timestamp]
        returns = returns_df['close'].pct_change().fillna(0)
        cumulative_returns = (((1 + returns).cumprod() - 1) * side_multiplier) - trade_cost
        df_filtered.loc[df_filtered['timestamp'] >= start_timestamp, 'net_pnl_pct'] = cumulative_returns
        df_filtered.loc[df_filtered['timestamp'] >= start_timestamp, 'filled_amount_quote'] = float(config.amount) * entry_price
        df_filtered['net_pnl_quote'] = df_filtered['net_pnl_pct'] * df_filtered['filled_amount_quote']
        df_filtered['cum_fees_quote'] = trade_cost * df_filtered['filled_amount_quote']

        # Make sure the trailing stop pct rises linearly to the net p/l pct when above the trailing stop trigger pct (if any)
        if trailing_sl_trigger_pct is not None and trailing_sl_delta_pct is not None:
            df_filtered.loc[(df_filtered['net_pnl_pct'] > trailing_sl_trigger_pct).cummax(), 'ts'] = (
                df_filtered['net_pnl_pct'] - float(trailing_sl_delta_pct)
            ).cummax()

        # Determine the earliest close event
        first_tp_timestamp = df_filtered[df_filtered['net_pnl_pct'] > tp]['timestamp'].min() if tp else None
        first_sl_timestamp = None
        if config.triple_barrier_config.stop_loss:
            sl = float(config.triple_barrier_config.stop_loss)
            sl_price = entry_price * (1 - sl * side_multiplier)
            sl_condition = df_filtered['low'] <= sl_price if config.side == TradeType.BUY else df_filtered['high'] >= sl_price
            first_sl_timestamp = df_filtered[sl_condition]['timestamp'].min()
        first_trailing_sl_timestamp = df_filtered[(~df_filtered['ts'].isna()) & (df_filtered['net_pnl_pct'] < df_filtered['ts'])]['timestamp'].min() if trailing_sl_delta_pct and trailing_sl_trigger_pct else None
        close_timestamp = min([timestamp for timestamp in [first_tp_timestamp, first_sl_timestamp, tl_timestamp, first_trailing_sl_timestamp] if not pd.isna(timestamp)])

        # Determine the close type
        if close_timestamp == first_tp_timestamp:
            close_type = CloseType.TAKE_PROFIT
        elif close_timestamp == first_sl_timestamp:
            close_type = CloseType.STOP_LOSS
        elif close_timestamp == first_trailing_sl_timestamp:
            close_type = CloseType.TRAILING_STOP
        else:
            close_type = CloseType.TIME_LIMIT

        # Set the final state of the DataFrame
        df_filtered = df_filtered[df_filtered['timestamp'] <= close_timestamp]
        df_filtered.loc[df_filtered.index[-1], "filled_amount_quote"] = df_filtered["filled_amount_quote"].iloc[-1] * 2

        # Construct and return ExecutorSimulation object
        simulation = ExecutorSimulation(
            config=config,
            executor_simulation=df_filtered,
            close_type=close_type
        )
        return simulation



import binascii
import hashlib
import hmac
import logging
import secrets
import textwrap
from typing import Dict

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_constants import BASE_URL, USER_AGENT
from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_web_utils import endpoint_from_url
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSJSONRequest, WSRequest
from hummingbot.logger import HummingbotLogger


class CoinbaseAdvancedTradeAuth(AuthBase):
    """
    Authentication class for Coinbase Advanced Trade API.

    Uses HMAC SHA256 to authenticate REST and websocket requests.

    Coinbase API documentation: https://docs.cdp.coinbase.com/sign-in-with-coinbase/docs/api-key-authentication
    """
    TIME_SYNC_UPDATE_S: float = 30
    _time_sync_last_updated_s: float = -1

    _logger: HummingbotLogger | logging.Logger | None = None

    @classmethod
    def logger(cls) -> HummingbotLogger | logging.Logger:
        if cls._logger is None:
            cls._logger = logging.getLogger(HummingbotLogger.logger_name_for_class(cls))
        return cls._logger

    __slots__ = (
        'api_key',
        'secret_key',
        'time_provider',
    )

    def __init__(self, api_key: str, secret_key: str, time_provider: TimeSynchronizer):
        """
        :param api_key: The API key.
        :param secret_key: The API secret key.
        :param time_provider: The time provider object.
        """
        self.api_key: str = api_key
        self.secret_key: str = secret_key
        self.time_provider: TimeSynchronizer = time_provider

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        This method is intended to configure a REST request to be authenticated.
        :param request: the request to be configured for authenticated interaction
        """
        try:
            return await self.rest_jwt_authenticate(request)
        except ValueError:
            self.logger().debug("Failed to authenticate using JWT. Attempting to authenticate using legacy method.")
            return await self.rest_legacy_authenticate(request)

    async def rest_legacy_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.

        All REST requests must contain the following headers:

        CB-ACCESS-KEY API key as a string
        CB-ACCESS-SIGN Message signature (see below)
        CB-ACCESS-TIMESTAMP Timestamp for your request
        All request bodies should have content type application/json and be valid JSON.

        Example request:

        curl https://api.coinbase.com/v2/user \
          --header "CB-ACCESS-KEY: <your api key>" \
          --header "CB-ACCESS-SIGN: <the user generated message signature>" \
          --header "CB-ACCESS-TIMESTAMP: <a timestamp for your request>"

        :param request: the request to be configured for authenticated interaction
        :returns: the authenticated request
        """
        timestamp: str = str(int(self.time_provider.time()))

        endpoint: str = endpoint_from_url(request.url).split('?')[0]  # ex: /v3/orders
        message = timestamp + str(request.method) + endpoint + str(request.data or '')
        signature: str = self._generate_signature(message=message)

        headers: Dict = dict(request.headers or {}) | {
            "accept": 'application/json',
            "content-type": 'application/json',
            "CB-ACCESS-KEY": self.api_key,
            "CB-ACCESS-SIGN": signature,
            "CB-ACCESS-TIMESTAMP": timestamp,
        }
        request.headers = headers

        return request

    async def rest_jwt_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the JWT header to the rest request.

        All JWT REST requests must contain the JWT Authorization in headers:

        Example request:
        curl https://api.coinbase.com/v2/user --header "Authorization: Bearer $JWT"

        :param request: the request to be configured for authenticated interaction
        :returns: the authenticated request
        """
        endpoint: str = endpoint_from_url(request.url).split('?')[0]  # ex: /v3/orders
        jwt_uri = f"{request.method} {BASE_URL}{endpoint}"

        try:
            token = self._build_jwt(jwt_uri)
            headers: Dict = dict(request.headers or {}) | {
                "content-type": 'application/json',
                "Authorization": f"Bearer {token}",
                "User-Agent": USER_AGENT,
            }
        except ValueError as e:
            raise e

        request.headers = headers

        return request

    async def ws_authenticate(self, request: WSJSONRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated.
        :param request: the request to be configured for authenticated interaction
        """

        try:
            return await self.ws_jwt_authenticate(request)
        except ValueError:
            self.logger().debug("Failed to authenticate using JWT. Attempting to authenticate using legacy method.")
            return await self.ws_legacy_authenticate(request)

    async def ws_legacy_authenticate(self, request: WSJSONRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated.
        :param request: the request to be configured for authenticated interaction
        https://docs.cdp.coinbase.com/advanced-trade/docs/ws-overview
        {
            "type": "subscribe",
            "product_ids": [
                "ETH-USD",
                "ETH-EUR"
            ],
            "channel": "level2",
            "api_key": "exampleApiKey123",
            "timestamp": 1660838876,
            "signature": "00000000000000000000000000",
        }
        To subscribe to any channel, users must provide a channel name, api_key, timestamp, and signature:
            channel name as a string. You can only subscribe to one channel at a time.
            timestamp should be a string in UNIX format. Example: "1677527973".
            signature should be created by:
        Concatenating and comma-separating the timestamp, channel name, and product Ids, for example: 1660838876level2ETH-USD,ETH-EUR.
        Signing the above message with the passphrase and base64-encoding the signature.
        """
        timestamp: str = str(int(self.time_provider.time()))

        products: str = ",".join(request.payload["product_ids"])
        message: str = timestamp + str(request.payload["channel"]) + products
        signature: str = self._generate_signature(message=message)

        payload: Dict = dict(request.payload or {}) | {
            "api_key": self.api_key,
            "signature": signature,
            "timestamp": timestamp,
        }
        request.payload = payload
        self.logger().debug(f"ws_authenticate payload: {payload}")

        return request

    async def ws_jwt_authenticate(self, request: WSJSONRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated.
        :param request: the request to be configured for authenticated interaction
        https://docs.cdp.coinbase.com/advanced-trade/docs/ws-overview
        {
            "type": "subscribe",
            "product_ids": [
                "ETH-USD",
                "ETH-EUR"
            ],
            "channel": "level2",
            "jwt": "exampleJWT",
        }
        To subscribe to any channel, users must provide a channel name, api_key, timestamp, and signature:
            channel name as a string. You can only subscribe to one channel at a time.
            timestamp should be a string in UNIX format. Example: "1677527973".
            signature should be created by:
        Concatenating and comma-separating the timestamp, channel name, and product Ids,f
        for example: 1660838876level2ETH-USD,ETH-EUR.
        Signing the above message with the passphrase and base64-encoding the signature.
        """
        try:
            payload: Dict = dict(request.payload or {}) | {
                "jwt": self._build_jwt(),
            }
            request.payload = payload
        except ValueError as e:
            raise e

        self.logger().debug(f"ws_authenticate payload: {payload}")

        return request

    def _generate_signature(self, *, message: str) -> str:
        """
        Generates an HMAC SHA256 signature from a message and the API secret key.

        :param message: the message to sign
        :returns: the signature
        """
        digest: str = hmac.new(self.secret_key.encode("utf8"), message.encode("utf8"), hashlib.sha256).digest().hex()
        return digest

    def _build_jwt(self, uri=None) -> str:
        """
        This is extracted from Coinbase SDK because it relies upon 'time' rather than the eim synchronizer
        """
        try:
            private_key_bytes = self._secret_key_pem().encode("utf-8")
            private_key = serialization.load_pem_private_key(
                private_key_bytes, password=None
            )
        except ValueError as e:
            # This handles errors like incorrect key format
            self.logger().debug("The API key is not PEM format. Falling back to Legacy sign-in.")
            raise e

        time_: int = int(self.time_provider.time())
        jwt_data = {
            "sub": self.api_key,
            "iss": "cdp",
            "nbf": time_,
            "exp": time_ + 120,
        }

        if uri is not None:
            jwt_data["uri"] = uri

        jwt_token = jwt.encode(
            jwt_data,
            private_key,
            algorithm="ES256",
            headers={"kid": self.api_key, "nonce": secrets.token_hex()},
        )
        self.logger().debug(f"JWT token: {jwt_token}")

        return jwt_token

    def _secret_key_pem(self) -> str:
        """
        Converts the secret key to PEM format.
        Comprehends keys in PEM format, single-line PEM and base64-encoded keys.
        """
        # Remove any leading or trailing whitespace and replace \\n with \n
        private_key_base64 = self.secret_key.strip().replace("\\n", "\n")

        # If the key is already in PEM format with \n, return it
        if private_key_base64.startswith("-----") and "\n" in private_key_base64:
            try:
                # Try to load the key to validate its structure
                serialization.load_pem_private_key(
                    private_key_base64.encode(),
                    password=None,
                    backend=default_backend()
                )
            except ValueError:
                raise ValueError("The secret key is not a valid PEM key.")
            return private_key_base64

        # Remove the BEGIN and END lines
        private_key_base64 = private_key_base64.replace("-----BEGIN EC PRIVATE" + " KEY-----", "").replace("-----END EC PRIVATE" + " KEY-----", "").strip()

        # Verify that the key is a correct base64 string
        try:
            binascii.a2b_base64(private_key_base64)
        except binascii.Error:
            raise ValueError("The secret key is not a valid base64 string.")

        # Wrap the base64-encoded key into lines of 64 characters
        wrapped_key = textwrap.wrap(private_key_base64, width=64)

        private_key_base64 = (
            "-----BEGIN" + " EC " + "PRIVATE" + " KEY-----\n"
            + "\n".join(wrapped_key)
            + "\n-----END" + " EC " + "PRIVATE" + " KEY-----"
        )

        try:
            # Try to load the key to validate its structure
            serialization.load_pem_private_key(
                private_key_base64.encode(),
                password=None,
                backend=default_backend()
            )
        except ValueError:
            raise ValueError("The secret key is not a valid PEM key.")
        return private_key_base64



import re
from typing import Callable, Dict, NamedTuple, Optional, Tuple

import hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_constants as constants
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.connector.utils import TimeSynchronizerRESTPreProcessor
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


def public_rest_url(path_url: str, domain: str = constants.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided public REST endpoint
    :param path_url: a public REST endpoint
    :param domain: the Coinbase Advanced Trade domain to connect to ("com" or "us"). The default value is "com"
    :return: the full URL to the endpoint
    """
    if any((path_url.startswith(p[:4]) for p in constants.SIGNIN_ENDPOINTS)):
        return constants.SIGNIN_URL.format(domain=domain) + path_url

    return constants.REST_URL.format(domain=domain) + path_url


def private_rest_url(path_url: str, domain: str = constants.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided private REST endpoint
    :param path_url: a private REST endpoint
    :param domain: the coinbase_advanced_trade domain to connect to ("com" or "us"). The default value is "com"
    :return: the full URL to the endpoint
    """
    if any((path_url.startswith(p) for p in constants.SIGNIN_ENDPOINTS)):
        return constants.SIGNIN_URL.format(domain=domain) + path_url

    return constants.REST_URL.format(domain=domain) + path_url


def endpoint_from_url(path_url: str, domain: str = constants.DEFAULT_DOMAIN) -> str:
    """
    Recreates the endpoint from the url
    :param path_url: URL to the endpoint
    :param domain: the coinbase_advanced_trade domain to connect to ("com" or "us"). The default value is "com"
    :return: the full URL to the endpoint
    """
    if domain not in path_url:
        raise ValueError(f"The domain {domain} is not part of the provided URL {path_url}")

    endpoint: str = re.split(domain, path_url)[1]

    # Must start with '/'
    if not endpoint.startswith("/"):
        endpoint = f"/{endpoint}"

    return endpoint


def build_api_factory(
        throttler: Optional[AsyncThrottler] = None,
        time_synchronizer: Optional[TimeSynchronizer] = None,
        domain: str = constants.DEFAULT_DOMAIN,
        time_provider: Optional[Callable] = None,
        auth: Optional[AuthBase] = None, ) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    time_synchronizer = time_synchronizer or TimeSynchronizer()
    time_provider = time_provider or (lambda: get_current_server_time_ms(
        throttler=throttler,
        domain=domain,
    ))
    return WebAssistantsFactory(
        throttler=throttler,
        auth=auth,
        rest_pre_processors=[
            TimeSynchronizerRESTPreProcessor(
                synchronizer=time_synchronizer, time_provider=time_provider
            ),
        ],
    )


def build_api_factory_without_time_synchronizer_pre_processor(throttler: AsyncThrottler) -> WebAssistantsFactory:
    return WebAssistantsFactory(throttler=throttler)


def create_throttler() -> AsyncThrottler:
    return AsyncThrottler(constants.RATE_LIMITS)


async def get_current_server_time_s(
        throttler: Optional[AsyncThrottler] = None,
        domain: str = constants.DEFAULT_DOMAIN,
) -> float:
    """
    Get the current server time in seconds
    :param throttler: the throttler to use for the request
    :param domain: the coinbase_advanced_trade domain to connect to ("com" or "us"). The default value is "com"
    :return: the current server time in seconds
    """
    """
    https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getservertime
    """
    throttler = throttler or create_throttler()
    api_factory = build_api_factory_without_time_synchronizer_pre_processor(throttler=throttler)
    rest_assistant = await api_factory.get_rest_assistant()
    response: Dict = await rest_assistant.execute_request(
        url=private_rest_url(path_url=constants.SERVER_TIME_EP, domain=domain),
        method=RESTMethod.GET,
        throttler_limit_id=constants.SERVER_TIME_EP,
    )
    server_time: float = float(get_timestamp_from_exchange_time(response["iso"], "s"))

    return server_time


async def get_current_server_time(
        throttler: Optional[AsyncThrottler] = None,
        domain: str = constants.DEFAULT_DOMAIN,
) -> float:
    return await get_current_server_time_s(throttler=throttler, domain=domain)


async def get_current_server_time_ms(
        throttler: Optional[AsyncThrottler] = None,
        domain: str = constants.DEFAULT_DOMAIN,
) -> int:
    server_time_s = await get_current_server_time_s(throttler=throttler, domain=domain)
    return int(server_time_s * 1000)


def get_timestamp_from_exchange_time(exchange_time: str, unit: str) -> float:
    assert isinstance(exchange_time, str), f"exchange_time should be str, not {type(exchange_time)}: {exchange_time}"
    assert isinstance(unit, str), f"unit should be str, not {type(unit)}"

    from dateutil import parser

    dt = parser.parse(timestr=exchange_time)
    t_s: float = dt.timestamp()
    if not isinstance(t_s, float):
        raise ValueError(f"Failed to parse exchange time {exchange_time}:{t_s}")
    if unit in {"s", "second", "seconds"}:
        return t_s
    elif unit in {"ms", "millisecond", "milliseconds"}:
        return t_s * 1000
    else:
        raise ValueError(f"Unsupported time unit {unit}")


def set_exchange_time_from_timestamp(timestamp: int | float, timestamp_unit: str = "s") -> str:
    assert isinstance(timestamp, (int, float)), f"timestamp should be int or float, not {type(timestamp)}"
    assert isinstance(timestamp_unit, str), f"timestamp_unit should be str, not {type(timestamp_unit)}"

    if timestamp_unit in {"ms", "millisecond", "milliseconds"}:
        timestamp /= 1000
    elif timestamp_unit not in ("s", "second", "seconds"):
        raise ValueError(f"Unsupported timestamp unit {timestamp_unit}")

    import datetime
    return f"{datetime.datetime.fromtimestamp(timestamp, datetime.UTC).isoformat()}"


def pair_to_symbol(pair: str) -> str:
    return pair


def symbol_to_pair(symbol: str) -> str:
    return symbol


class CoinbaseAdvancedTradeWSSMessage(NamedTuple):
    """
    Coinbase Advanced Trade Websocket API message
    https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels
    ```json
    {
      "channel": "market_trades",
      "client_id": "",
      "timestamp": "2023-02-09T20:19:35.39625135Z",
      "sequence_num": 0,
      "events": [
        ...
      ]
    }
    ```
    """

    channel: str
    client_id: str
    timestamp: str
    sequence_num: int
    events: Tuple


class CoinbaseAdvancedTradeServerIssueException(Exception):
    """
    Exception raised when the Coinbase Advanced Trade server returns an error
    """
    pass



import typing
from collections import defaultdict
from dataclasses import dataclass, field
from decimal import Decimal
from typing import Dict, List, Optional

from hummingbot.connector.utils import combine_to_hb_trading_pair, split_hb_trading_pair
from hummingbot.core.data_type.common import OrderType, PositionAction, TradeType
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.utils.estimate_fee import build_perpetual_trade_fee, build_trade_fee

if typing.TYPE_CHECKING:  # avoid circular import problems
    from hummingbot.connector.exchange_base import ExchangeBase


@dataclass
class OrderCandidate:
    """
    WARNING: Do not use this class for sizing. Instead, use the `BudgetChecker`.

    This class contains a full picture of the impact of a potential order on the user account.

    It can return a dictionary with the base collateral required for an order, the percentage-fee collateral
    and the fixed-fee collaterals, and any combination of those. In addition, it contains a field sizing
    the potential return of an order.

    It also provides logic to adjust the order size, the collateral values, and the return based on
    a dictionary of currently available assets in the user account.
    """
    trading_pair: str
    is_maker: bool
    order_type: OrderType
    order_side: TradeType
    amount: Decimal
    price: Decimal
    order_collateral: Optional[TokenAmount] = field(default=None, init=False)
    percent_fee_collateral: Optional[TokenAmount] = field(default=None, init=False)
    percent_fee_value: Optional[TokenAmount] = field(default=None, init=False)
    fixed_fee_collaterals: List[TokenAmount] = field(default=list, init=False)
    potential_returns: Optional[TokenAmount] = field(default=None, init=False)
    resized: bool = field(default=False, init=False)
    from_total_balances: bool = False

    @property
    def collateral_dict(self) -> Dict:
        cd = defaultdict(lambda: Decimal("0"))
        if self.order_collateral is not None:
            cd[self.order_collateral.token] += self.order_collateral.amount
        if self.percent_fee_collateral is not None:
            cd[self.percent_fee_collateral.token] += self.percent_fee_collateral.amount
        for entry in self.fixed_fee_collaterals:
            cd[entry.token] += entry.amount
        return cd

    @property
    def is_zero_order(self) -> bool:
        return self.amount == Decimal("0")

    def get_size_token_and_order_size(self) -> TokenAmount:
        trading_pair = self.trading_pair
        base, quote = split_hb_trading_pair(trading_pair)
        if self.order_side == TradeType.BUY:
            order_size = self.amount * self.price
            size_token = quote
        else:
            order_size = self.amount
            size_token = base
        return TokenAmount(size_token, order_size)

    def set_to_zero(self):
        self._scale_order(scaler=Decimal("0"))

    def populate_collateral_entries(self, exchange: 'ExchangeBase'):
        self._populate_order_collateral_entry(exchange)
        fee = self._get_fee(exchange)
        self._populate_percent_fee_collateral_entry(exchange, fee)
        self._populate_fixed_fee_collateral_entries(fee)
        self._populate_potential_returns_entry(exchange)
        self._populate_percent_fee_value(exchange, fee)
        self._apply_fee_impact_on_potential_returns(exchange, fee)

    def adjust_from_balances(self, available_balances: Dict[str, Decimal]):
        if not self.is_zero_order:
            self._adjust_for_order_collateral(available_balances)
        if not self.is_zero_order:
            self._adjust_for_percent_fee_collateral(available_balances)
        if not self.is_zero_order:
            self._adjust_for_fixed_fee_collaterals(available_balances)

    def _populate_order_collateral_entry(self, exchange: 'ExchangeBase'):
        oc_token = self._get_order_collateral_token(exchange)
        if oc_token is not None:
            oc_amount = self._get_order_collateral_amount(exchange, oc_token)
            self.order_collateral = TokenAmount(oc_token, oc_amount)

    def _get_order_collateral_token(self, exchange: 'ExchangeBase') -> Optional[str]:
        trading_pair = self.trading_pair
        base, quote = split_hb_trading_pair(trading_pair)
        if self.order_side == TradeType.BUY:
            oc_token = quote
        else:
            oc_token = base
        return oc_token

    def _get_order_collateral_amount(
        self, exchange: 'ExchangeBase', order_collateral_token: str
    ) -> Decimal:
        size_token, order_size = self.get_size_token_and_order_size()
        size_collateral_price = self._get_size_collateral_price(exchange, order_collateral_token)
        oc_amount = order_size * size_collateral_price
        return oc_amount

    def _populate_percent_fee_collateral_entry(self, exchange: 'ExchangeBase', fee: TradeFeeBase):
        impact = fee.get_fee_impact_on_order_cost(self, exchange)
        if impact is not None:
            token, amount = impact
            self.percent_fee_collateral = TokenAmount(token, amount)

    def _populate_fixed_fee_collateral_entries(self, fee: TradeFeeBase):
        self.fixed_fee_collaterals = []
        for token, amount in fee.flat_fees:
            self.fixed_fee_collaterals.append(
                TokenAmount(token, amount))

    def _populate_potential_returns_entry(self, exchange: 'ExchangeBase'):
        r_token = self._get_returns_token(exchange)
        if r_token is not None:
            r_amount = self._get_returns_amount(exchange)
            self.potential_returns = TokenAmount(r_token, r_amount)

    def _populate_percent_fee_value(self, exchange: 'ExchangeBase', fee: TradeFeeBase):
        cost_impact = fee.get_fee_impact_on_order_cost(self, exchange)
        if cost_impact is not None:
            self.percent_fee_value = cost_impact
        else:
            returns_impact = fee.get_fee_impact_on_order_returns(self, exchange)
            if returns_impact is not None:
                impact_token = self.potential_returns.token
                self.percent_fee_value = TokenAmount(impact_token, returns_impact)

    def _apply_fee_impact_on_potential_returns(self, exchange: 'ExchangeBase', fee: TradeFeeBase):
        if self.potential_returns is not None:
            impact = fee.get_fee_impact_on_order_returns(self, exchange)
            if impact is not None:
                self.potential_returns.amount -= impact

    def _get_returns_token(self, exchange: 'ExchangeBase') -> Optional[str]:
        trading_pair = self.trading_pair
        base, quote = split_hb_trading_pair(trading_pair)
        if self.order_side == TradeType.BUY:
            r_token = base
        else:
            r_token = quote
        return r_token

    def _get_returns_amount(self, exchange: 'ExchangeBase') -> Decimal:
        if self.order_side == TradeType.BUY:
            r_amount = self.amount
        else:
            r_amount = self.amount * self.price
        return r_amount

    def _get_size_collateral_price(
        self, exchange: 'ExchangeBase', order_collateral_token: str
    ) -> Decimal:
        size_token, _ = self.get_size_token_and_order_size()
        base, quote = split_hb_trading_pair(self.trading_pair)

        if order_collateral_token == size_token:
            price = Decimal("1")
        elif order_collateral_token == base:  # size_token == quote
            price = Decimal("1") / self.price
        elif order_collateral_token == quote:  # size_token == base
            price = self.price
        else:
            size_collateral_pair = combine_to_hb_trading_pair(size_token, order_collateral_token)
            price = exchange.get_price(size_collateral_pair, is_buy=True)  # we are buying

        return price

    def _adjust_for_order_collateral(self, available_balances: Dict[str, Decimal]):
        if self.order_collateral is not None:
            token, amount = self.order_collateral
            if not amount.is_nan() and available_balances[token] < amount:
                scaler = available_balances[token] / amount
                self._scale_order(scaler)

    def _adjust_for_percent_fee_collateral(self, available_balances: Dict[str, Decimal]):
        if self.percent_fee_collateral is not None:
            token, amount = self.percent_fee_collateral
            if token == self.order_collateral.token:
                amount += self.order_collateral.amount
            if available_balances[token] < amount:
                scaler = available_balances[token] / amount
                self._scale_order(scaler)

    def _adjust_for_fixed_fee_collaterals(self, available_balances: Dict[str, Decimal]):
        oc_token = self.order_collateral.token if self.order_collateral is not None else None
        pfc_token = self.percent_fee_collateral.token if self.percent_fee_collateral is not None else None
        oc_amount, pfc_amount = self._get_order_and_pf_collateral_amounts_for_ff_adjustment()

        for collateral_entry in self.fixed_fee_collaterals:
            ffc_token, ffc_amount = collateral_entry
            available_balance = available_balances[ffc_token]
            if available_balance < ffc_amount:
                self._scale_order(scaler=Decimal("0"))
                break
            if oc_token is not None and ffc_token == oc_token and available_balance < ffc_amount + oc_amount:
                scaler = (available_balance - ffc_amount) / oc_amount
                self._scale_order(scaler)
                oc_amount, pfc_amount = self._get_order_and_pf_collateral_amounts_for_ff_adjustment()
            if pfc_token is not None and ffc_token == pfc_token and available_balance < ffc_amount + pfc_amount:
                scaler = (available_balance - ffc_amount) / pfc_amount
                self._scale_order(scaler)
                oc_amount, pfc_amount = self._get_order_and_pf_collateral_amounts_for_ff_adjustment()
            if self.is_zero_order:
                break

    def _get_order_and_pf_collateral_amounts_for_ff_adjustment(self) -> TokenAmount:
        if self.order_collateral is not None:
            oc_token, oc_amount = self.order_collateral
        else:
            oc_token = None
            oc_amount = Decimal("0")
        if self.percent_fee_collateral is not None:
            pfc_token, pfc_amount = self.percent_fee_collateral
            if oc_token is not None and pfc_token == oc_token:
                oc_amount += pfc_amount
                pfc_amount = Decimal("0")
        else:
            pfc_amount = Decimal("0")
        return TokenAmount(oc_amount, pfc_amount)

    def _get_fee(self, exchange: 'ExchangeBase') -> TradeFeeBase:
        trading_pair = self.trading_pair
        price = self.price
        base, quote = split_hb_trading_pair(trading_pair)
        fee = build_trade_fee(
            exchange.name,
            self.is_maker,
            base,
            quote,
            self.order_type,
            self.order_side,
            self.amount,
            price,
        )

        return fee

    def _scale_order(self, scaler: Decimal):
        self.amount *= scaler
        if self.order_collateral is not None:
            self.order_collateral.amount *= scaler
        if self.percent_fee_collateral is not None:
            self.percent_fee_collateral.amount *= scaler
        if self.percent_fee_value is not None:
            self.percent_fee_value.amount *= scaler
        if self.potential_returns is not None:
            self.potential_returns.amount *= scaler
        if self.is_zero_order:
            self.order_collateral = None
            self.percent_fee_collateral = None
            self.percent_fee_value = None
            self.fixed_fee_collaterals = []
            self.potential_returns = None
        self.resized = True


@dataclass
class PerpetualOrderCandidate(OrderCandidate):
    leverage: Decimal = Decimal("1")
    position_close: bool = False

    def _get_order_collateral_token(self, exchange: 'ExchangeBase') -> Optional[str]:
        if self.position_close:
            oc_token = None  # the contract is the collateral
        else:
            oc_token = self._get_collateral_token(exchange)
        return oc_token

    def _get_order_collateral_amount(
        self, exchange: 'ExchangeBase', order_collateral_token: str
    ) -> Decimal:
        if self.position_close:
            oc_amount = Decimal("0")  # the contract is the collateral
        else:
            oc_amount = self._get_collateral_amount(exchange)
        return oc_amount

    def _populate_percent_fee_collateral_entry(self, exchange: 'ExchangeBase', fee: TradeFeeBase):
        if not self.position_close:
            super()._populate_percent_fee_collateral_entry(exchange, fee)
            if (
                self.percent_fee_collateral is not None
                and self.percent_fee_collateral.token == self.order_collateral.token
            ):
                leverage = self.leverage
                self.percent_fee_collateral.amount *= leverage

    def _populate_percent_fee_value(self, exchange: 'ExchangeBase', fee: TradeFeeBase):
        if not self.position_close:
            super()._populate_percent_fee_value(exchange, fee)
            if (
                self.percent_fee_value is not None
                and self.percent_fee_value.token == self.order_collateral.token
            ):
                leverage = self.leverage
                self.percent_fee_value.amount *= leverage

    def _get_returns_token(self, exchange: 'ExchangeBase') -> Optional[str]:
        if self.position_close:
            r_token = self._get_collateral_token(exchange)
        else:
            r_token = None  # the contract is the returns
        return r_token

    def _get_returns_amount(self, exchange: 'ExchangeBase') -> Decimal:
        if self.position_close:
            r_amount = self._get_collateral_amount(exchange)
        else:
            r_amount = Decimal("0")  # the contract is the returns
        return r_amount

    def _get_collateral_amount(self, exchange: 'ExchangeBase') -> Decimal:
        if self.position_close:
            self._flip_order_side()
        size_token, order_size = self.get_size_token_and_order_size()
        if self.position_close:
            self._flip_order_side()
        order_token = self._get_collateral_token(exchange)
        size_collateral_price = self._get_size_collateral_price(exchange, order_token)
        amount = order_size * size_collateral_price / self.leverage
        return amount

    def _get_collateral_token(self, exchange: 'ExchangeBase') -> str:
        trading_pair = self.trading_pair
        if self.order_side == TradeType.BUY:
            token = exchange.get_buy_collateral_token(trading_pair)
        else:
            token = exchange.get_sell_collateral_token(trading_pair)
        return token

    def _flip_order_side(self):
        self.order_side = (
            TradeType.BUY if self.order_side == TradeType.SELL
            else TradeType.SELL
        )

    def _get_fee(self, exchange: 'ExchangeBase') -> TradeFeeBase:
        base, quote = split_hb_trading_pair(self.trading_pair)
        position_action = PositionAction.CLOSE if self.position_close else PositionAction.OPEN
        fee = build_perpetual_trade_fee(
            exchange.name,
            self.is_maker,
            position_action,
            base,
            quote,
            self.order_type,
            self.order_side,
            self.amount,
            self.price,
        )

        return fee



from dataclasses import dataclass
from decimal import Decimal

from pyinjective.core.market import DerivativeMarket, SpotMarket
from pyinjective.core.token import Token

from hummingbot.connector.utils import combine_to_hb_trading_pair


@dataclass(frozen=True)
class InjectiveToken:
    unique_symbol: str
    native_token: Token

    @property
    def denom(self) -> str:
        return self.native_token.denom

    @property
    def symbol(self) -> str:
        return self.native_token.symbol

    @property
    def name(self) -> str:
        return self.native_token.name

    @property
    def decimals(self) -> int:
        return self.native_token.decimals

    def value_from_chain_format(self, chain_value: Decimal) -> Decimal:
        scaler = Decimal(f"1e{-self.decimals}")
        return chain_value * scaler

    def value_from_special_chain_format(self, chain_value: Decimal) -> Decimal:
        scaler = Decimal(f"1e{-self.decimals-18}")
        return chain_value * scaler


@dataclass(frozen=True)
class InjectiveSpotMarket:
    market_id: str
    base_token: InjectiveToken
    quote_token: InjectiveToken
    native_market: SpotMarket

    def trading_pair(self):
        return combine_to_hb_trading_pair(self.base_token.unique_symbol, self.quote_token.unique_symbol)

    def quantity_from_chain_format(self, chain_quantity: Decimal) -> Decimal:
        return self.base_token.value_from_chain_format(chain_value=chain_quantity)

    def price_from_chain_format(self, chain_price: Decimal) -> Decimal:
        scaler = Decimal(f"1e{self.base_token.decimals-self.quote_token.decimals}")
        return chain_price * scaler

    def quantity_from_special_chain_format(self, chain_quantity: Decimal) -> Decimal:
        quantity = chain_quantity / Decimal("1e18")
        return self.quantity_from_chain_format(chain_quantity=quantity)

    def price_from_special_chain_format(self, chain_price: Decimal) -> Decimal:
        price = chain_price / Decimal("1e18")
        return self.price_from_chain_format(chain_price=price)

    def min_price_tick_size(self) -> Decimal:
        return self.price_from_chain_format(chain_price=self.native_market.min_price_tick_size)

    def min_quantity_tick_size(self) -> Decimal:
        return self.quantity_from_chain_format(chain_quantity=self.native_market.min_quantity_tick_size)

    def maker_fee_rate(self) -> Decimal:
        return self.native_market.maker_fee_rate

    def taker_fee_rate(self) -> Decimal:
        return self.native_market.taker_fee_rate

    def min_notional(self) -> Decimal:
        return self.quote_token.value_from_chain_format(chain_value=self.native_market.min_notional)


@dataclass(frozen=True)
class InjectiveDerivativeMarket:
    market_id: str
    quote_token: InjectiveToken
    native_market: DerivativeMarket

    def base_token_symbol(self):
        ticker_base, _ = self.native_market.ticker.split("/") if "/" in self.native_market.ticker else (self.native_market.ticker, 0)
        return ticker_base

    def trading_pair(self):
        ticker_base, _ = self.native_market.ticker.split("/") if "/" in self.native_market.ticker else (self.native_market.ticker, 0)
        return combine_to_hb_trading_pair(ticker_base, self.quote_token.unique_symbol)

    def quantity_from_chain_format(self, chain_quantity: Decimal) -> Decimal:
        return chain_quantity

    def price_from_chain_format(self, chain_price: Decimal) -> Decimal:
        scaler = Decimal(f"1e{-self.quote_token.decimals}")
        return chain_price * scaler

    def quantity_from_special_chain_format(self, chain_quantity: Decimal) -> Decimal:
        quantity = chain_quantity / Decimal("1e18")
        return self.quantity_from_chain_format(chain_quantity=quantity)

    def price_from_special_chain_format(self, chain_price: Decimal) -> Decimal:
        price = chain_price / Decimal("1e18")
        return self.price_from_chain_format(chain_price=price)

    def min_price_tick_size(self) -> Decimal:
        return self.price_from_chain_format(chain_price=self.native_market.min_price_tick_size)

    def min_quantity_tick_size(self) -> Decimal:
        return self.quantity_from_chain_format(chain_quantity=self.native_market.min_quantity_tick_size)

    def maker_fee_rate(self) -> Decimal:
        return self.native_market.maker_fee_rate

    def taker_fee_rate(self) -> Decimal:
        return self.native_market.taker_fee_rate

    def oracle_base(self) -> str:
        return self.native_market.oracle_base

    def oracle_quote(self) -> str:
        return self.native_market.oracle_quote

    def oracle_type(self) -> str:
        return self.native_market.oracle_type

    def min_notional(self) -> Decimal:
        return self.quote_token.value_from_chain_format(chain_value=self.native_market.min_notional)



import asyncio
import importlib
import inspect
import os
from decimal import Decimal
from typing import Callable, Dict, List, Optional, Set

import pandas as pd
import yaml
from pydantic import field_validator
from pydantic.v1 import Field

from hummingbot.client import settings
from hummingbot.client.config.config_data_types import BaseClientModel, ClientFieldData
from hummingbot.client.ui.interface_utils import format_df_for_printout
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.connector.markets_recorder import MarketsRecorder
from hummingbot.core.data_type.common import PositionMode
from hummingbot.data_feed.candles_feed.data_types import CandlesConfig
from hummingbot.data_feed.market_data_provider import MarketDataProvider
from hummingbot.exceptions import InvalidController
from hummingbot.strategy.script_strategy_base import ScriptStrategyBase
from hummingbot.strategy_v2.controllers.controller_base import ControllerBase, ControllerConfigBase
from hummingbot.strategy_v2.controllers.directional_trading_controller_base import (
    DirectionalTradingControllerConfigBase,
)
from hummingbot.strategy_v2.controllers.market_making_controller_base import MarketMakingControllerConfigBase
from hummingbot.strategy_v2.executors.executor_orchestrator import ExecutorOrchestrator
from hummingbot.strategy_v2.models.base import RunnableStatus
from hummingbot.strategy_v2.models.executor_actions import (
    CreateExecutorAction,
    ExecutorAction,
    StopExecutorAction,
    StoreExecutorAction,
)
from hummingbot.strategy_v2.models.executors_info import ExecutorInfo


class StrategyV2ConfigBase(BaseClientModel):
    """
    Base class for version 2 strategy configurations.
    """
    markets: Dict[str, Set[str]] = Field(
        default="binance_perpetual.JASMY-USDT,RLC-USDT",
        client_data=ClientFieldData(
            prompt_on_new=True,
            prompt=lambda mi: (
                "Enter markets in format 'exchange1.tp1,tp2:exchange2.tp1,tp2':"
            )
        )
    )
    candles_config: List[CandlesConfig] = Field(
        default="binance_perpetual.JASMY-USDT.1m.500:binance_perpetual.RLC-USDT.1m.500",
        client_data=ClientFieldData(
            prompt_on_new=True,
            prompt=lambda mi: (
                "Enter candle configs in format 'exchange1.tp1.interval1.max_records:"
                "exchange2.tp2.interval2.max_records':"
            )
        )
    )
    controllers_config: List[str] = Field(
        default=None,
        client_data=ClientFieldData(
            is_updatable=True,
            prompt_on_new=True,
            prompt=lambda mi: "Enter controller configurations (comma-separated file paths), leave it empty if none: "
        ))
    config_update_interval: Optional[int] = Field(
        default=60,
        gt=0,
        client_data=ClientFieldData(
            prompt_on_new=False,
            prompt=lambda mi: "Enter the config update interval in seconds (e.g. 60): ",
        )
    )

    @field_validator("controllers_config", mode="before")
    @classmethod
    def parse_controllers_config(cls, v):
        # Parse string input into a list of file pathsq
        if isinstance(v, str):
            if v == "":
                return []
            return [item.strip() for item in v.split(',') if item.strip()]
        if v is None:
            return []
        return v

    def load_controller_configs(self):
        loaded_configs = []
        for config_path in self.controllers_config:
            full_path = os.path.join(settings.CONTROLLERS_CONF_DIR_PATH, config_path)
            with open(full_path, 'r') as file:
                config_data = yaml.safe_load(file)

            controller_type = config_data.get('controller_type')
            controller_name = config_data.get('controller_name')

            if not controller_type or not controller_name:
                raise ValueError(f"Missing controller_type or controller_name in {config_path}")

            module_path = f"{settings.CONTROLLERS_MODULE}.{controller_type}.{controller_name}"
            module = importlib.import_module(module_path)

            config_class = next((member for member_name, member in inspect.getmembers(module)
                                 if inspect.isclass(member) and member not in [ControllerConfigBase,
                                                                               MarketMakingControllerConfigBase,
                                                                               DirectionalTradingControllerConfigBase]
                                 and (issubclass(member, ControllerConfigBase))), None)
            if not config_class:
                raise InvalidController(f"No configuration class found in the module {controller_name}.")

            loaded_configs.append(config_class(**config_data))

        return loaded_configs

    @field_validator('markets', mode="before")
    @classmethod
    def parse_markets(cls, v) -> Dict[str, Set[str]]:
        if isinstance(v, str):
            return cls.parse_markets_str(v)
        elif isinstance(v, dict):
            return v
        raise ValueError("Invalid type for markets. Expected str or Dict[str, Set[str]]")

    @staticmethod
    def parse_markets_str(v: str) -> Dict[str, Set[str]]:
        markets_dict = {}
        if v.strip():
            exchanges = v.split(':')
            for exchange in exchanges:
                parts = exchange.split('.')
                if len(parts) != 2 or not parts[1]:
                    raise ValueError(f"Invalid market format in segment '{exchange}'. "
                                     "Expected format: 'exchange.tp1,tp2'")
                exchange_name, trading_pairs = parts
                markets_dict[exchange_name] = set(trading_pairs.split(','))
        return markets_dict

    @field_validator('candles_config', mode="before")
    @classmethod
    def parse_candles_config(cls, v) -> List[CandlesConfig]:
        if isinstance(v, str):
            return cls.parse_candles_config_str(v)
        elif isinstance(v, list):
            return v
        raise ValueError("Invalid type for candles_config. Expected str or List[CandlesConfig]")

    @staticmethod
    def parse_candles_config_str(v: str) -> List[CandlesConfig]:
        configs = []
        if v.strip():
            entries = v.split(':')
            for entry in entries:
                parts = entry.split('.')
                if len(parts) != 4:
                    raise ValueError(f"Invalid candles config format in segment '{entry}'. "
                                     "Expected format: 'exchange.tradingpair.interval.maxrecords'")
                connector, trading_pair, interval, max_records_str = parts
                try:
                    max_records = int(max_records_str)
                except ValueError:
                    raise ValueError(f"Invalid max_records value '{max_records_str}' in segment '{entry}'. "
                                     "max_records should be an integer.")
                config = CandlesConfig(
                    connector=connector,
                    trading_pair=trading_pair,
                    interval=interval,
                    max_records=max_records
                )
                configs.append(config)
        return configs


class StrategyV2Base(ScriptStrategyBase):
    """
    V2StrategyBase is a base class for strategies that use the new smart components architecture.
    """
    markets: Dict[str, Set[str]]
    _last_config_update_ts: float = 0
    closed_executors_buffer: int = 100
    max_executors_close_attempts: int = 10

    @classmethod
    def init_markets(cls, config: StrategyV2ConfigBase):
        """
        Initialize the markets that the strategy is going to use. This method is called when the strategy is created in
        the start command. Can be overridden to implement custom behavior.
        """
        markets = config.markets
        controllers_configs = config.load_controller_configs()
        for controller_config in controllers_configs:
            markets = controller_config.update_markets(markets)
        cls.markets = markets

    def __init__(self, connectors: Dict[str, ConnectorBase], config: Optional[StrategyV2ConfigBase] = None):
        super().__init__(connectors, config)
        # Initialize the executor orchestrator
        self.config = config
        self.executor_orchestrator = ExecutorOrchestrator(strategy=self)

        self.executors_info: Dict[str, List[ExecutorInfo]] = {}
        self.positions_held: Dict[str, List] = {}

        # Create a queue to listen to actions from the controllers
        self.actions_queue = asyncio.Queue()
        self.listen_to_executor_actions_task: asyncio.Task = asyncio.create_task(self.listen_to_executor_actions())

        # Initialize the market data provider
        self.market_data_provider = MarketDataProvider(connectors)
        self.market_data_provider.initialize_candles_feed_list(config.candles_config)
        self.controllers: Dict[str, ControllerBase] = {}
        self.initialize_controllers()
        self._is_stop_triggered = False

    def initialize_controllers(self):
        """
        Initialize the controllers based on the provided configuration.
        """
        controllers_configs = self.config.load_controller_configs()
        for controller_config in controllers_configs:
            self.add_controller(controller_config)
            MarketsRecorder.get_instance().store_controller_config(controller_config)

    def add_controller(self, config: ControllerConfigBase):
        try:
            controller = config.get_controller_class()(config, self.market_data_provider, self.actions_queue)
            controller.start()
            self.controllers[config.id] = controller
        except Exception as e:
            self.logger().error(f"Error adding controller: {e}", exc_info=True)

    def update_controllers_configs(self):
        """
        Update the controllers configurations based on the provided configuration.
        """
        if self._last_config_update_ts + self.config.config_update_interval < self.current_timestamp:
            self._last_config_update_ts = self.current_timestamp
            controllers_configs = self.config.load_controller_configs()
            for controller_config in controllers_configs:
                if controller_config.id in self.controllers:
                    self.controllers[controller_config.id].update_config(controller_config)
                else:
                    self.add_controller(controller_config)

    async def listen_to_executor_actions(self):
        """
        Asynchronously listen to actions from the controllers and execute them.
        """
        while True:
            try:
                actions = await self.actions_queue.get()
                self.executor_orchestrator.execute_actions(actions)
                self.update_executors_info()
                controller_id = actions[0].controller_id
                controller = self.controllers.get(controller_id)
                controller.executors_info = self.executors_info.get(controller_id, [])
                controller.executors_update_event.set()
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.logger().error(f"Error executing action: {e}", exc_info=True)

    def update_executors_info(self):
        """
        Update the local state of the executors and publish the updates to the active controllers.
        In this case we are going to update the controllers directly with the executors info so the event is not
        set and is managed with the async queue.
        """
        try:
            self.executors_info = self.executor_orchestrator.get_executors_report()
            self.positions_held = self.executor_orchestrator.get_positions_report()
            for controllers in self.controllers.values():
                controllers.executors_info = self.executors_info.get(controllers.config.id, [])
                controllers.positions_held = self.positions_held.get(controllers.config.id, [])
        except Exception as e:
            self.logger().error(f"Error updating executors info: {e}", exc_info=True)

    @staticmethod
    def is_perpetual(connector: str) -> bool:
        return "perpetual" in connector

    async def on_stop(self):
        self._is_stop_triggered = True
        self.executor_orchestrator.stop()
        self.market_data_provider.stop()
        self.listen_to_executor_actions_task.cancel()
        for controller in self.controllers.values():
            controller.stop()
        for i in range(self.max_executors_close_attempts):
            if all([executor.is_done for executor in self.get_all_executors()]):
                continue
            await asyncio.sleep(5.0)
        self.executor_orchestrator.store_all_executors()

    def on_tick(self):
        self.update_executors_info()
        self.update_controllers_configs()
        if self.market_data_provider.ready and not self._is_stop_triggered:
            executor_actions: List[ExecutorAction] = self.determine_executor_actions()
            for action in executor_actions:
                self.executor_orchestrator.execute_action(action)

    def determine_executor_actions(self) -> List[ExecutorAction]:
        """
        Determine actions based on the provided executor handler report.
        """
        actions = []
        actions.extend(self.create_actions_proposal())
        actions.extend(self.stop_actions_proposal())
        actions.extend(self.store_actions_proposal())
        return actions

    def create_actions_proposal(self) -> List[CreateExecutorAction]:
        """
        Create actions proposal based on the current state of the executors.
        """
        raise NotImplementedError

    def stop_actions_proposal(self) -> List[StopExecutorAction]:
        """
        Create a list of actions to stop the executors based on order refresh and early stop conditions.
        """
        raise NotImplementedError

    def store_actions_proposal(self) -> List[StoreExecutorAction]:
        """
        Create a list of actions to store the executors that have been stopped.
        """
        potential_executors_to_store = self.filter_executors(
            executors=self.get_all_executors(),
            filter_func=lambda x: x.is_done)
        sorted_executors = sorted(potential_executors_to_store, key=lambda x: x.timestamp, reverse=True)
        if len(sorted_executors) > self.closed_executors_buffer:
            return [StoreExecutorAction(executor_id=executor.id, controller_id=executor.controller_id) for executor in
                    sorted_executors[self.closed_executors_buffer:]]
        return []

    def get_executors_by_controller(self, controller_id: str) -> List[ExecutorInfo]:
        return self.executors_info.get(controller_id, [])

    def get_all_executors(self) -> List[ExecutorInfo]:
        return [executor for executors in self.executors_info.values() for executor in executors]

    def set_leverage(self, connector: str, trading_pair: str, leverage: int):
        self.connectors[connector].set_leverage(trading_pair, leverage)

    def set_position_mode(self, connector: str, position_mode: PositionMode):
        self.connectors[connector].set_position_mode(position_mode)

    @staticmethod
    def filter_executors(executors: List[ExecutorInfo], filter_func: Callable[[ExecutorInfo], bool]) -> List[ExecutorInfo]:
        return [executor for executor in executors if filter_func(executor)]

    @staticmethod
    def executors_info_to_df(executors_info: List[ExecutorInfo]) -> pd.DataFrame:
        """
        Convert a list of executor handler info to a dataframe.
        """
        df = pd.DataFrame([ei.to_dict() for ei in executors_info])
        # Convert the enum values to integers
        df['status'] = df['status'].apply(lambda x: x.value)

        # Sort the DataFrame
        df.sort_values(by='status', ascending=True, inplace=True)

        # Convert back to enums for display
        df['status'] = df['status'].apply(RunnableStatus)
        return df

    def format_status(self) -> str:
        if not self.ready_to_trade:
            return "Market connectors are not ready."
        lines = []
        warning_lines = []
        warning_lines.extend(self.network_warning(self.get_market_trading_pair_tuples()))

        balance_df = self.get_balance_df()
        lines.extend(["", "  Balances:"] + ["    " + line for line in balance_df.to_string(index=False).split("\n")])

        try:
            df = self.active_orders_df()
            lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
        except ValueError:
            lines.extend(["", "  No active maker orders."])
        columns_to_show = ["type", "side", "status", "net_pnl_pct", "net_pnl_quote", "cum_fees_quote",
                           "filled_amount_quote", "is_trading", "close_type", "age"]

        # Initialize global performance metrics
        global_realized_pnl_quote = Decimal(0)
        global_unrealized_pnl_quote = Decimal(0)
        global_volume_traded = Decimal(0)
        global_close_type_counts = {}

        # Process each controller
        for controller_id, controller in self.controllers.items():
            lines.append(f"\n\nController: {controller_id}")
            # Append controller market data metrics
            lines.extend(controller.to_format_status())
            # executors_list = self.get_executors_by_controller(controller_id)
            # if len(executors_list) == 0:
            #     lines.append("No executors found.")
            # else:
            #     # In memory executors info
            #     executors_df = self.executors_info_to_df(executors_list)
            #     executors_df["age"] = self.current_timestamp - executors_df["timestamp"]
            #     lines.extend([format_df_for_printout(executors_df[columns_to_show], table_format="psql")])

            # Generate performance report for each controller
            performance_report = self.executor_orchestrator.generate_performance_report(controller_id)

            # Append performance metrics
            controller_performance_info = [
                f"Realized PNL (Quote): {performance_report.realized_pnl_quote:.2f} | Unrealized PNL (Quote): {performance_report.unrealized_pnl_quote:.2f}"
                f"--> Global PNL (Quote): {performance_report.global_pnl_quote:.2f} | Global PNL (%): {performance_report.global_pnl_pct:.2f}%",
                f"Total Volume Traded: {performance_report.volume_traded:.2f}"
            ]

            # Add position summary if available
            if hasattr(performance_report, "positions_summary") and performance_report.positions_summary:
                controller_performance_info.append("\nPositions Held Summary:")
                controller_performance_info.append("-" * 170)
                controller_performance_info.append(
                    f"{'Connector':<20} | "
                    f"{'Trading Pair':<12} | "
                    f"{'Side':<4} | "
                    f"{'Volume':<12} | "
                    f"{'Units':<10} | "
                    f"{'Value (USD)':<12} | "
                    f"{'BEP':<16} | "
                    f"{'Realized PNL':<12} | "
                    f"{'Unreal. PNL':<12} | "
                    f"{'Fees':<10} | "
                    f"{'Global PNL':<12}"
                )
                controller_performance_info.append("-" * 170)
                for pos in performance_report.positions_summary:
                    controller_performance_info.append(
                        f"{pos.connector_name:<20} | "
                        f"{pos.trading_pair:<12} | "
                        f"{pos.side.name:<4} | "
                        f"${pos.volume_traded_quote:>11.2f} | "
                        f"{pos.amount:>10.4f} | "
                        f"${pos.amount * pos.breakeven_price:<11.2f} | "
                        f"{pos.breakeven_price:>16.6f} | "
                        f"${pos.realized_pnl_quote:>+11.2f} | "
                        f"${pos.unrealized_pnl_quote:>+11.2f} | "
                        f"${pos.cum_fees_quote:>9.2f} | "
                        f"${pos.global_pnl_quote:>10.2f}"
                    )
                controller_performance_info.append("-" * 170)

            # Append close type counts
            if performance_report.close_type_counts:
                controller_performance_info.append("Close Types Count:")
                for close_type, count in performance_report.close_type_counts.items():
                    controller_performance_info.append(f"  {close_type}: {count}")
            lines.extend(controller_performance_info)

            # Aggregate global metrics and close type counts
            global_realized_pnl_quote += performance_report.realized_pnl_quote
            global_unrealized_pnl_quote += performance_report.unrealized_pnl_quote
            global_volume_traded += performance_report.volume_traded
            for close_type, value in performance_report.close_type_counts.items():
                global_close_type_counts[close_type] = global_close_type_counts.get(close_type, 0) + value

        main_executors_list = self.get_executors_by_controller("main")
        if len(main_executors_list) > 0:
            lines.append("\n\nMain Controller Executors:")
            main_executors_df = self.executors_info_to_df(main_executors_list)
            main_executors_df["age"] = self.current_timestamp - main_executors_df["timestamp"]
            lines.extend([format_df_for_printout(main_executors_df[columns_to_show], table_format="psql")])
            main_performance_report = self.executor_orchestrator.generate_performance_report("main")
            # Aggregate global metrics and close type counts
            global_realized_pnl_quote += main_performance_report.realized_pnl_quote
            global_unrealized_pnl_quote += main_performance_report.unrealized_pnl_quote
            global_volume_traded += main_performance_report.volume_traded
            for close_type, value in main_performance_report.close_type_counts.items():
                global_close_type_counts[close_type] = global_close_type_counts.get(close_type, 0) + value

        # Calculate and append global performance metrics
        global_pnl_quote = global_realized_pnl_quote + global_unrealized_pnl_quote
        global_pnl_pct = (global_pnl_quote / global_volume_traded) * 100 if global_volume_traded != 0 else Decimal(0)

        global_performance_summary = [
            "\n\nGlobal Performance Summary:",
            f"Global PNL (Quote): {global_pnl_quote:.2f} | Global PNL (%): {global_pnl_pct:.2f}% | Total Volume Traded (Global): {global_volume_traded:.2f}"
        ]

        # Append global close type counts
        if global_close_type_counts:
            global_performance_summary.append("Global Close Types Count:")
            for close_type, count in global_close_type_counts.items():
                global_performance_summary.append(f"  {close_type}: {count}")

        lines.extend(global_performance_summary)

        # Combine original and extra information
        return "\n".join(lines)



from typing import Callable

from pydantic.v1.json import pydantic_encoder

from hummingbot.client.config.config_var import ConfigVar


def new_fee_config_var(key: str, type_str: str = "decimal"):
    return ConfigVar(key=key,
                     prompt=None,
                     required_if=lambda: False,
                     type_str=type_str)


def using_exchange(exchange: str) -> Callable:
    from hummingbot.client.settings import required_exchanges
    return lambda: exchange in required_exchanges


def strategy_config_schema_encoder(o):
    if callable(o):
        return None
    else:
        return pydantic_encoder(o)



from decimal import Decimal
from enum import Enum
from typing import List

from hummingbot.core.utils.async_utils import safe_gather
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple

from .data_types import ArbProposal, ArbProposalSide, TokenAmount

s_decimal_nan = Decimal("NaN")


class TradeDirection(Enum):
    BUY = 1
    SELL = 0


async def create_arb_proposals(
        market_info_1: MarketTradingPairTuple,
        market_info_2: MarketTradingPairTuple,
        market_1_extra_flat_fees: List[TokenAmount],
        market_2_extra_flat_fees: List[TokenAmount],
        order_amount: Decimal
) -> List[ArbProposal]:
    order_amount = Decimal(str(order_amount))
    results = []

    tasks = []
    for trade_direction in TradeDirection:
        is_buy = trade_direction == TradeDirection.BUY
        tasks.append([
            market_info_1.market.get_quote_price(market_info_1.trading_pair, is_buy, order_amount),
            market_info_1.market.get_order_price(market_info_1.trading_pair, is_buy, order_amount),
            market_info_2.market.get_quote_price(market_info_2.trading_pair, not is_buy, order_amount),
            market_info_2.market.get_order_price(market_info_2.trading_pair, not is_buy, order_amount)
        ])

    results_raw = await safe_gather(*[safe_gather(*task_group) for task_group in tasks])

    for trade_direction, task_group_result in zip(TradeDirection, results_raw):
        is_buy = trade_direction == TradeDirection.BUY
        m_1_q_price, m_1_o_price, m_2_q_price, m_2_o_price = task_group_result

        if any(p is None for p in (m_1_o_price, m_1_q_price, m_2_o_price, m_2_q_price)):
            continue

        first_side = ArbProposalSide(
            market_info=market_info_1,
            is_buy=is_buy,
            quote_price=m_1_q_price,
            order_price=m_1_o_price,
            amount=order_amount,
            extra_flat_fees=market_1_extra_flat_fees,
        )
        second_side = ArbProposalSide(
            market_info=market_info_2,
            is_buy=not is_buy,
            quote_price=m_2_q_price,
            order_price=m_2_o_price,
            amount=order_amount,
            extra_flat_fees=market_2_extra_flat_fees
        )

        results.append(ArbProposal(first_side, second_side))

    return results



import json
import time
from collections import OrderedDict

import eth_account
import msgpack
from eth_account.messages import encode_typed_data
from eth_utils import keccak, to_hex

from hummingbot.connector.exchange.hyperliquid import hyperliquid_constants as CONSTANTS
from hummingbot.connector.exchange.hyperliquid.hyperliquid_web_utils import order_spec_to_order_wire
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class HyperliquidAuth(AuthBase):
    """
    Auth class required by Hyperliquid API
    """

    def __init__(self, api_key: str, api_secret: str, use_vault: bool):
        self._api_key: str = api_key
        self._api_secret: str = api_secret
        self._use_vault: bool = use_vault
        self.wallet = eth_account.Account.from_key(api_secret)

    @classmethod
    def address_to_bytes(cls, address):
        return bytes.fromhex(address[2:] if address.startswith("0x") else address)

    @classmethod
    def action_hash(cls, action, vault_address, nonce):
        data = msgpack.packb(action)
        data += nonce.to_bytes(8, "big")
        if vault_address is None:
            data += b"\x00"
        else:
            data += b"\x01"
            data += cls.address_to_bytes(vault_address)
        return keccak(data)

    def sign_inner(self, wallet, data):
        structured_data = encode_typed_data(full_message=data)
        signed = wallet.sign_message(structured_data)
        return {"r": to_hex(signed["r"]), "s": to_hex(signed["s"]), "v": signed["v"]}

    def construct_phantom_agent(self, hash, is_mainnet):
        return {"source": "a" if is_mainnet else "b", "connectionId": hash}

    def sign_l1_action(self, wallet, action, active_pool, nonce, is_mainnet):
        _hash = self.action_hash(action, active_pool, nonce)
        phantom_agent = self.construct_phantom_agent(_hash, is_mainnet)

        data = {
            "domain": {
                "chainId": 1337,
                "name": "Exchange",
                "verifyingContract": "0x0000000000000000000000000000000000000000",
                "version": "1",
            },
            "types": {
                "Agent": [
                    {"name": "source", "type": "string"},
                    {"name": "connectionId", "type": "bytes32"},
                ],
                "EIP712Domain": [
                    {"name": "name", "type": "string"},
                    {"name": "version", "type": "string"},
                    {"name": "chainId", "type": "uint256"},
                    {"name": "verifyingContract", "type": "address"},
                ],
            },
            "primaryType": "Agent",
            "message": phantom_agent,
        }
        return self.sign_inner(wallet, data)

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        base_url = request.url
        if request.method == RESTMethod.POST:
            request.data = self.add_auth_to_params_post(request.data, base_url)
        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        return request  # pass-through

    def _sign_update_leverage_params(self, params, base_url, timestamp):
        signature = self.sign_l1_action(
            self.wallet,
            params,
            None if not self._use_vault else self._api_key,
            timestamp,
            CONSTANTS.BASE_URL in base_url,
        )
        payload = {
            "action": params,
            "nonce": timestamp,
            "signature": signature,
            "vaultAddress": self._api_key if self._use_vault else None,
        }
        return payload

    def _sign_cancel_params(self, params, base_url, timestamp):
        order_action = {
            "type": "cancelByCloid",
            "cancels": [params["cancels"]],
        }
        signature = self.sign_l1_action(
            self.wallet,
            order_action,
            None if not self._use_vault else self._api_key,
            timestamp,
            CONSTANTS.BASE_URL in base_url,
        )
        payload = {
            "action": order_action,
            "nonce": timestamp,
            "signature": signature,
            "vaultAddress": self._api_key if self._use_vault else None,

        }
        return payload

    def _sign_order_params(self, params, base_url, timestamp):

        order = params["orders"]
        grouping = params["grouping"]
        order_action = {
            "type": "order",
            "orders": [order_spec_to_order_wire(order)],
            "grouping": grouping,
        }
        signature = self.sign_l1_action(
            self.wallet,
            order_action,
            None if not self._use_vault else self._api_key,
            timestamp,
            CONSTANTS.BASE_URL in base_url,
        )

        payload = {
            "action": order_action,
            "nonce": timestamp,
            "signature": signature,
            "vaultAddress": self._api_key if self._use_vault else None,

        }
        return payload

    def add_auth_to_params_post(self, params: str, base_url):
        timestamp = int(self._get_timestamp() * 1e3)
        payload = {}
        data = json.loads(params) if params is not None else {}

        request_params = OrderedDict(data or {})

        request_type = request_params.get("type")
        if request_type == "order":
            payload = self._sign_order_params(request_params, base_url, timestamp)
        elif request_type == "cancel":
            payload = self._sign_cancel_params(request_params, base_url, timestamp)
        elif request_type == "updateLeverage":
            payload = self._sign_update_leverage_params(request_params, base_url, timestamp)
        payload = json.dumps(payload)
        return payload

    @staticmethod
    def _get_timestamp():
        return time.time()



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Callable, Optional

from pydantic import BaseModel, ConfigDict, field_validator
from pydantic.v1 import Field

from hummingbot.client.config.config_validators import validate_connector, validate_decimal


class ClientConfigEnum(Enum):
    def __str__(self):
        return self.value


@dataclass()
class ClientFieldData:
    prompt: Optional[Callable[['BaseClientModel'], str]] = None
    prompt_on_new: bool = False
    is_secure: bool = False
    is_connect_key: bool = False
    is_updatable: bool = False


class BaseClientModel(BaseModel):
    model_config = ConfigDict(validate_assignment=True, title=None, extra="forbid", json_encoders={
        datetime: lambda dt: dt.strftime("%Y-%m-%d %H:%M:%S"),
    })

    @classmethod
    def _clear_schema_cache(cls):
        cls.__schema_cache__ = {}

    def is_required(self, attr: str) -> bool:
        default = self.model_fields[attr].default
        if self.model_fields[attr].annotation._name != "Optional" and (default == {} or default.default == Ellipsis):
            return True
        else:
            return False

    def validate_decimal(v: str, field: Field):
        """Used for client-friendly error output."""
        field_info = field.field_info
        inclusive = field_info.ge is not None or field_info.le is not None
        min_value = field_info.gt if field_info.gt is not None else field_info.ge
        min_value = Decimal(min_value) if min_value is not None else min_value
        max_value = field_info.lt if field_info.lt is not None else field_info.le
        max_value = Decimal(max_value) if max_value is not None else max_value
        ret = validate_decimal(v, min_value, max_value, inclusive)
        if ret is not None:
            raise ValueError(ret)
        return v


class BaseConnectorConfigMap(BaseClientModel):
    connector: str = Field(
        default=...,
        description="What is your connector?",
        client_data=ClientFieldData(
            prompt=lambda mi: "What is your connector?",
            prompt_on_new=True,
        ),
    )

    @field_validator("connector", mode="before")
    @classmethod
    def validate_connector(cls, v: str):
        ret = validate_connector(v)
        if ret is not None:
            raise ValueError(ret)
        return v



import asyncio
import time
from collections import defaultdict
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional

import hummingbot.connector.derivative.hyperliquid_perpetual.hyperliquid_perpetual_constants as CONSTANTS
import hummingbot.connector.derivative.hyperliquid_perpetual.hyperliquid_perpetual_web_utils as web_utils
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.perpetual_api_order_book_data_source import PerpetualAPIOrderBookDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.derivative.hyperliquid_perpetual.hyperliquid_perpetual_derivative import (
        HyperliquidPerpetualDerivative,
    )


class HyperliquidPerpetualAPIOrderBookDataSource(PerpetualAPIOrderBookDataSource):
    _bpobds_logger: Optional[HummingbotLogger] = None
    _trading_pair_symbol_map: Dict[str, Mapping[str, str]] = {}
    _mapping_initialization_lock = asyncio.Lock()

    def __init__(
            self,
            trading_pairs: List[str],
            connector: 'HyperliquidPerpetualDerivative',
            api_factory: WebAssistantsFactory,
            domain: str = CONSTANTS.DOMAIN
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._domain = domain
        self._trading_pairs: List[str] = trading_pairs
        self._message_queue: Dict[str, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._snapshot_messages_queue_key = "order_book_snapshot"

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def get_funding_info(self, trading_pair: str) -> FundingInfo:
        response: List = await self._request_complete_funding_info(trading_pair)
        ex_trading_pair = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        coin = ex_trading_pair.split("-")[0]
        for index, i in enumerate(response[0]['universe']):
            if i['name'] == coin:
                funding_info = FundingInfo(
                    trading_pair=trading_pair,
                    index_price=Decimal(response[1][index]['oraclePx']),
                    mark_price=Decimal(response[1][index]['markPx']),
                    next_funding_utc_timestamp=self._next_funding_time(),
                    rate=Decimal(response[1][index]['funding']),
                )
                return funding_info

    async def listen_for_funding_info(self, output: asyncio.Queue):
        """
        Reads the funding info events queue and updates the local funding info information.
        """
        while True:
            try:
                for trading_pair in self._trading_pairs:
                    funding_info = await self.get_funding_info(trading_pair)
                    funding_info_update = FundingInfoUpdate(
                        trading_pair=trading_pair,
                        index_price=funding_info.index_price,
                        mark_price=funding_info.mark_price,
                        next_funding_utc_timestamp=funding_info.next_funding_utc_timestamp,
                        rate=funding_info.rate,
                    )
                    output.put_nowait(funding_info_update)
                await self._sleep(CONSTANTS.FUNDING_RATE_UPDATE_INTERNAL_SECOND)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error when processing public funding info updates from exchange")
                await self._sleep(CONSTANTS.FUNDING_RATE_UPDATE_INTERNAL_SECOND)

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        ex_trading_pair = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        coin = ex_trading_pair.split("-")[0]
        params = {
            "type": 'l2Book',
            "coin": coin
        }

        data = await self._connector._api_post(
            path_url=CONSTANTS.SNAPSHOT_REST_URL,
            data=params)
        return data

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_response.update({"trading_pair": trading_pair})
        snapshot_msg: OrderBookMessage = OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": snapshot_response["trading_pair"],
            "update_id": int(snapshot_response['time']),
            "bids": [[float(i['px']), float(i['sz'])] for i in snapshot_response['levels'][0]],
            "asks": [[float(i['px']), float(i['sz'])] for i in snapshot_response['levels'][1]],
        }, timestamp=int(snapshot_response['time']))
        return snapshot_msg

    async def _connected_websocket_assistant(self) -> WSAssistant:
        url = f"{web_utils.wss_url(self._domain)}"
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.

        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            for trading_pair in self._trading_pairs:
                symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                coin = symbol.split("-")[0]
                trades_payload = {
                    "method": "subscribe",
                    "subscription": {
                        "type": CONSTANTS.TRADES_ENDPOINT_NAME,
                        "coin": coin,
                    }
                }
                subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trades_payload)

                order_book_payload = {
                    "method": "subscribe",
                    "subscription": {
                        "type": CONSTANTS.DEPTH_ENDPOINT_NAME,
                        "coin": coin,
                    }
                }
                subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=order_book_payload)

                await ws.send(subscribe_trade_request)
                await ws.send(subscribe_orderbook_request)

                self.logger().info("Subscribed to public order book, trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error("Unexpected error occurred subscribing to order book data streams.")
            raise

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if "result" not in event_message:
            stream_name = event_message.get("channel")
            if "l2Book" in stream_name:
                channel = self._snapshot_messages_queue_key
            elif "trades" in stream_name:
                channel = self._trade_messages_queue_key
        return channel

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        timestamp: float = raw_message["data"]["time"] * 1e-3
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(
            raw_message["data"]["coin"] + '-' + CONSTANTS.CURRENCY)
        data = raw_message["data"]
        order_book_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.DIFF, {
            "trading_pair": trading_pair,
            "update_id": data["time"],
            "bids": [[float(i['px']), float(i['sz'])] for i in data["levels"][0]],
            "asks": [[float(i['px']), float(i['sz'])] for i in data["levels"][1]],
        }, timestamp=timestamp)
        message_queue.put_nowait(order_book_message)

    async def _parse_order_book_snapshot_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        timestamp: float = raw_message["data"]["time"] * 1e-3
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(
            raw_message["data"]["coin"] + '-' + CONSTANTS.CURRENCY)
        data = raw_message["data"]
        order_book_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": trading_pair,
            "update_id": data["time"],
            "bids": [[float(i['px']), float(i['sz'])] for i in data["levels"][0]],
            "asks": [[float(i['px']), float(i['sz'])] for i in data["levels"][1]],
        }, timestamp=timestamp)
        message_queue.put_nowait(order_book_message)

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        data = raw_message["data"]
        for trade_data in data:
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(
                trade_data["coin"] + '-' + CONSTANTS.CURRENCY)
            trade_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.TRADE, {
                "trading_pair": trading_pair,
                "trade_type": float(TradeType.SELL.value) if trade_data["side"] == "A" else float(
                    TradeType.BUY.value),
                "trade_id": trade_data["hash"],
                "price": float(trade_data["px"]),
                "amount": float(trade_data["sz"])
            }, timestamp=trade_data["time"] * 1e-3)

            message_queue.put_nowait(trade_message)

    async def _parse_funding_info_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        pass

    async def _request_complete_funding_info(self, trading_pair: str):
        data = await self._connector._api_post(path_url=CONSTANTS.EXCHANGE_INFO_URL,
                                               data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})
        return data

    def _next_funding_time(self) -> int:
        """
        Funding settlement occurs every 1 hours as mentioned in https://hyperliquid.gitbook.io/hyperliquid-docs/trading/funding
        """
        return int(((time.time() // 3600) + 1) * 3600)



from decimal import Decimal
from typing import (
    NamedTuple, Iterator
)
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_query_result import ClientOrderBookQueryResult
from hummingbot.core.data_type.order_book_row import ClientOrderBookRow
from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.core.data_type.common import PriceType


class MarketTradingPairTuple(NamedTuple):
    market: ExchangeBase
    trading_pair: str
    base_asset: str
    quote_asset: str

    def __repr__(self) -> str:
        return f"MarketTradingPairTuple({self.market.name}, {self.trading_pair}, {self.base_asset}, {self.quote_asset})"

    @property
    def order_book(self) -> OrderBook:
        return self.market.get_order_book(self.trading_pair)

    @property
    def quote_balance(self) -> Decimal:
        return self.market.get_balance(self.quote_asset)

    @property
    def base_balance(self) -> Decimal:
        return self.market.get_balance(self.base_asset)

    def get_mid_price(self) -> Decimal:
        return self.market.get_mid_price(self.trading_pair)

    def get_price(self, is_buy: bool) -> Decimal:
        return self.market.get_price(self.trading_pair, is_buy)

    def get_price_by_type(self, price_type: PriceType) -> Decimal:
        return self.market.get_price_by_type(self.trading_pair, price_type)

    def get_vwap_for_volume(self, is_buy: bool, volume: Decimal) -> ClientOrderBookQueryResult:
        return self.market.get_vwap_for_volume(self.trading_pair, is_buy, volume)

    def get_price_for_volume(self, is_buy: bool, volume: Decimal) -> ClientOrderBookQueryResult:
        return self.market.get_price_for_volume(self.trading_pair, is_buy, volume)

    def order_book_bid_entries(self) -> Iterator[ClientOrderBookRow]:
        return self.market.order_book_bid_entries(self.trading_pair)

    def order_book_ask_entries(self) -> Iterator[ClientOrderBookRow]:
        return self.market.order_book_ask_entries(self.trading_pair)



import base64
import hashlib
import hmac
import time
from typing import Any, Dict

import hummingbot.connector.exchange.btc_markets.btc_markets_constants as CONSTANTS
from hummingbot.connector.exchange.btc_markets import btc_markets_web_utils as web_utils
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSRequest


class BtcMarketsAuth(AuthBase):
    """
    Auth class required by btc_markets API
    Learn more at https://api.btcmarkets.net/doc/v3#section/Authentication/Authentication-process
    """
    def __init__(self, api_key: str, secret_key: str, time_provider: TimeSynchronizer):
        self.api_key = api_key
        self.secret_key = secret_key
        self.time_provider = time_provider

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """
        now = self._timestamp_in_milliseconds()
        sig = self.get_signature(
            request.method.name,
            web_utils.get_path_from_url(request.url),
            now,
            request.data if request.method.name == "POST" else {}
        )

        headers = self._generate_auth_headers(now, sig)
        if request.headers is not None:
            headers.update(request.headers)
        request.headers = headers

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. BtcMarkets does not use this
        functionality
        """
        return request  # pass-through

    def get_referral_code_headers(self):
        """
        Generates authentication headers required by BtcMarkets
        :return: a dictionary of auth headers
        """
        return {
            "referer": CONSTANTS.HBOT_BROKER_ID
        }

    def get_signature(
        self,
        method: str,
        path_url: str,
        nonce: int,
        data: Dict[str, Any] = None
    ):
        """
        Generates authentication signature and return it in a dictionary along with other inputs
        :return: a dictionary of request info including the request signature
        """
        data = data or {}

        if data is None or data == {}:
            payload = f"{method}/{path_url}{nonce}{''}"
        else:
            bjson = str(data)
            payload = f"{method}/{path_url}{nonce}{bjson}"

        return self._generate_signature(payload)

    def _generate_auth_headers(self, nonce: int, sig: str):
        """
        Generates HTTP headers
        """
        headers = {
            "Accept": "application/json",
            "Accept-Charset": "UTF-8",
            "Content-Type": "application/json",
            "BM-AUTH-APIKEY": self.api_key,
            "BM-AUTH-TIMESTAMP": str(nonce),
            "BM-AUTH-SIGNATURE": sig
        }

        return headers

    def _generate_signature(self, payload: str) -> str:
        """
        Generates a presigned signature
        :return: a signature of auth params
        """
        digest = base64.b64encode(hmac.new(
            base64.b64decode(self.secret_key), payload.encode("utf8"), digestmod=hashlib.sha512).digest())
        return digest.decode('utf8')

    def _generate_auth_dict_ws(self, nonce: int) -> str:
        """
        Generates an authentication params for websockets login
        :return: a signature of auth params
        """

        payload = "/users/self/subscribe" + "\n" + str(nonce)
        return self._generate_signature(payload)

    def generate_ws_authentication_message(self) -> str:
        """
        Generates the authentication message to start receiving messages from
        the 3 private ws channels
        """
        now = self._timestamp_in_milliseconds()
        signature = self._generate_auth_dict_ws(now)
        auth_message = {
            "signature": signature,
            "key": self.api_key,
            "timestamp": str(now),
            "messageType": CONSTANTS.SUBSCRIBE,
        }
        return auth_message

    def _timestamp_in_milliseconds(self) -> int:
        return int(self._time() * 1e3)

    def _time(self):
        return time.time()



import asyncio
import contextlib
import functools
import logging
import uuid
from collections import defaultdict, deque
from typing import Any, Dict, Optional, Tuple, Union
from unittest.mock import AsyncMock, PropertyMock

import aiohttp

from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.logger import HummingbotLogger


def get_stable_key(ws: AsyncMock) -> uuid.UUID:
    """
    Recursively unwraps the websocket mock to retrieve the original stable key.
    If no _stable_key is found, falls back to using the object's id.
    """
    current = ws
    while True:
        if hasattr(current, "_stable_key"):
            return current._stable_key
        elif hasattr(current, "__wrapped__"):
            current = current.__wrapped__
        else:
            return id(current)


class MockWebsocketClientSession:
    # Created this class instead of using a generic mock to be sure that no other methods from the client session
    # are required when working with websockets
    def __init__(self, mock_websocket: AsyncMock):
        self._mock_websocket = mock_websocket
        self._connection_args: Optional[Tuple[Any]] = None
        self._connection_kwargs: Optional[Dict[str, Any]] = None

    @property
    def connection_args(self) -> Tuple[Any]:
        return self._connection_args or ()

    @property
    def connection_kwargs(self) -> Dict[str, Any]:
        return self._connection_kwargs or {}

    async def ws_connect(self, *args, **kwargs):
        self._connection_args = args
        self._connection_kwargs = kwargs
        return self._mock_websocket

    async def close(self):
        pass


class NetworkMockingAssistant:
    _logger: HummingbotLogger | logging.Logger | None = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, deprecated_loop=None):
        super().__init__()
        self._response_text_queues: dict[uuid.UUID, asyncio.Queue] | None = None
        self._response_json_queues: dict[uuid.UUID, asyncio.Queue] | None = None
        self._response_status_queues: dict[uuid.UUID, deque] | None = None
        self._sent_http_requests: dict[uuid.UUID, asyncio.Queue] | None = None

        self._incoming_websocket_json_queues: dict[uuid.UUID, asyncio.Queue] | None = None
        self._all_incoming_websocket_json_delivered_event: dict[uuid.UUID, asyncio.Event] | None = None
        self._incoming_websocket_text_queues: dict[uuid.UUID, asyncio.Queue] | None = None
        self._all_incoming_websocket_text_delivered_event: dict[uuid.UUID, asyncio.Event] | None = None
        self._incoming_websocket_aiohttp_queues: dict[uuid.UUID, asyncio.Queue] | None = None
        self._all_incoming_websocket_aiohttp_delivered_event: dict[uuid.UUID, asyncio.Event] | None = None
        self._sent_websocket_json_messages: dict[uuid.UUID, list] | None = None
        self._sent_websocket_text_messages: dict[uuid.UUID, list] | None = None

        try:
            self._loop_id = id(asyncio.get_running_loop())
            asyncio.create_task(self.async_init())
        except RuntimeError:
            self._loop_id = None

    async def async_init(self):
        self._response_text_queues: dict[uuid.UUID, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._response_json_queues: dict[uuid.UUID, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._response_status_queues: dict[uuid.UUID, deque] = defaultdict(deque)
        self._sent_http_requests: dict[uuid.UUID, asyncio.Queue] = defaultdict(asyncio.Queue)

        self._incoming_websocket_json_queues: dict[uuid.UUID, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._all_incoming_websocket_json_delivered_event: dict[uuid.UUID, asyncio.Event] = defaultdict(asyncio.Event)
        self._incoming_websocket_text_queues: dict[uuid.UUID, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._all_incoming_websocket_text_delivered_event: dict[uuid.UUID, asyncio.Event] = defaultdict(asyncio.Event)
        self._incoming_websocket_aiohttp_queues: dict[uuid.UUID, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._all_incoming_websocket_aiohttp_delivered_event = defaultdict(asyncio.Event)
        self._sent_websocket_json_messages: dict[uuid.UUID, list] = defaultdict(list)
        self._sent_websocket_text_messages: dict[uuid.UUID, list] = defaultdict(list)

    def verify_async_init(self):
        if any(
                attr is None
                for attr in [
                    self._response_text_queues,
                    self._response_json_queues,
                    self._response_status_queues,
                    self._sent_http_requests,
                    self._incoming_websocket_json_queues,
                    self._all_incoming_websocket_json_delivered_event,
                    self._incoming_websocket_text_queues,
                    self._all_incoming_websocket_text_delivered_event,
                    self._incoming_websocket_aiohttp_queues,
                    self._all_incoming_websocket_aiohttp_delivered_event,
                    self._sent_websocket_json_messages,
                    self._sent_websocket_text_messages
                ]
        ):
            raise Exception("NetworkMockingAssistant must be initialized in async context. Please call async_init() first.")

        with contextlib.suppress(RuntimeError):
            if self._loop_id != id(asyncio.get_running_loop()):
                raise Exception("NetworkMockingAssistant was initialized on a different event loop.")

    @staticmethod
    def async_partial(function, *args, **kwargs):
        """
        Returns an async function that always calls `function` with the bound arguments,
        completely ignoring any extra parameters passed at call time.
        """

        async def partial_func(*_args, **_kwargs):
            return await function(*args, **kwargs)

        return partial_func

    def _get_next_api_response_status(self, http_mock):
        self.verify_async_init()
        return self._response_status_queues[http_mock].popleft()

    async def _get_next_api_response_json(self, http_mock):
        self.verify_async_init()
        return await self._response_json_queues[http_mock].get()

    async def _get_next_api_response_text(self, http_mock):
        self.verify_async_init()
        return await self._response_text_queues[http_mock].get()

    def _handle_http_request(self, http_mock, url, headers=None, params=None, data=None, *args, **kwargs):
        self.verify_async_init()
        response = AsyncMock()
        type(response).status = PropertyMock(side_effect=functools.partial(
            self._get_next_api_response_status, http_mock))
        response.json.side_effect = self.async_partial(self._get_next_api_response_json, http_mock)
        response.text.side_effect = self.async_partial(self._get_next_api_response_text, http_mock)
        response.__aenter__.return_value = response

        components = params if params else data
        self._sent_http_requests[http_mock].put_nowait((url, headers, components))

        return response

    def configure_web_assistants_factory(self, web_assistants_factory: WebAssistantsFactory) -> AsyncMock:
        websocket_mock: AsyncMock = self.create_websocket_mock()
        client_session_mock: MockWebsocketClientSession = MockWebsocketClientSession(mock_websocket=websocket_mock)

        web_assistants_factory._connections_factory._ws_independent_session = client_session_mock

        return websocket_mock

    def configure_http_request_mock(self, http_request_mock):
        http_request_mock.side_effect = functools.partial(self._handle_http_request, http_request_mock)

    def add_http_response(self, http_request_mock, response_status, response_json=None, response_text=None):
        self.verify_async_init()
        self._response_status_queues[http_request_mock].append(response_status)
        if response_json is not None:
            self._response_json_queues[http_request_mock].put_nowait(response_json)
        if response_text is not None:
            self._response_text_queues[http_request_mock].put_nowait(response_text)

    async def next_sent_request_data(self, http_request_mock):
        self.verify_async_init()
        return await self._sent_http_requests[http_request_mock].get()

    async def _get_next_websocket_json_message(self, ws_key: uuid.UUID, *args, **kwargs):
        self.verify_async_init()
        queue = self._incoming_websocket_json_queues[ws_key]
        message = await queue.get()
        if queue.empty():
            self._all_incoming_websocket_json_delivered_event[ws_key].set()
        return message

    async def _get_next_websocket_aiohttp_message(self, ws_key: uuid.UUID, *args, **kwargs):
        self.verify_async_init()
        queue = self._incoming_websocket_aiohttp_queues[ws_key]
        message = await queue.get()
        if queue.empty():
            self._all_incoming_websocket_aiohttp_delivered_event[ws_key].set()
        if isinstance(message, (BaseException, Exception)):
            raise message
        return message

    async def _get_next_websocket_text_message(self, ws_key: uuid.UUID, *args, **kwargs):
        self.verify_async_init()
        queue = self._incoming_websocket_text_queues[ws_key]
        message = await queue.get()
        if queue.empty():
            self._all_incoming_websocket_text_delivered_event[ws_key].set()
        return message

    def create_websocket_mock(self):
        self.verify_async_init()
        ws = AsyncMock()
        # Create a stable key
        stable_key: uuid.UUID = uuid.uuid4()
        ws._stable_key = stable_key
        # Ensure __aenter__ returns the same ws instance (with stable key intact)
        ws.__aenter__.side_effect = lambda: ws
        ws.__aexit__.return_value = None

        # Set side effects using async_partial with ignore_first_arg if needed.
        ws.send_json.side_effect = lambda sent_message: self._sent_websocket_json_messages[stable_key].append(
            sent_message)
        ws.send.side_effect = lambda sent_message: self._sent_websocket_text_messages[stable_key].append(sent_message)
        ws.send_str.side_effect = lambda sent_message: self._sent_websocket_text_messages[stable_key].append(
            sent_message)
        ws.receive_json.side_effect = self.async_partial(self._get_next_websocket_json_message, stable_key)
        ws.receive_str.side_effect = self.async_partial(self._get_next_websocket_text_message, stable_key)
        ws.receive.side_effect = self.async_partial(self._get_next_websocket_aiohttp_message, stable_key)
        ws.recv.side_effect = self.async_partial(self._get_next_websocket_text_message, stable_key)
        return ws

    def add_websocket_json_message(self, websocket_mock, message):
        self.verify_async_init()
        key: uuid.UUID = get_stable_key(websocket_mock)
        self._incoming_websocket_json_queues[key].put_nowait(message)
        self._all_incoming_websocket_json_delivered_event[key].clear()

    def add_websocket_text_message(self, websocket_mock, message):
        self.verify_async_init()
        key: uuid.UUID = get_stable_key(websocket_mock)
        self._incoming_websocket_text_queues[key].put_nowait(message)
        self._all_incoming_websocket_text_delivered_event[key].clear()

    def add_websocket_aiohttp_message(
            self,
            websocket_mock: AsyncMock,
            message: str,
            message_type: aiohttp.WSMsgType = aiohttp.WSMsgType.TEXT
    ):
        self.verify_async_init()
        key: uuid.UUID = get_stable_key(websocket_mock)
        msg = aiohttp.WSMessage(message_type, message, extra=None)
        self._incoming_websocket_aiohttp_queues[key].put_nowait(msg)
        self._all_incoming_websocket_aiohttp_delivered_event[key].clear()

    def add_websocket_aiohttp_exception(self, websocket_mock, exception: Union[Exception, BaseException]):
        self.verify_async_init()
        key: uuid.UUID = get_stable_key(websocket_mock)
        self._incoming_websocket_aiohttp_queues[key].put_nowait(exception)
        self._all_incoming_websocket_aiohttp_delivered_event[key].clear()

    def json_messages_sent_through_websocket(self, websocket_mock):
        self.verify_async_init()
        key: uuid.UUID = get_stable_key(websocket_mock)
        return self._sent_websocket_json_messages[key]

    def text_messages_sent_through_websocket(self, websocket_mock):
        self.verify_async_init()
        key: uuid.UUID = get_stable_key(websocket_mock)
        return self._sent_websocket_text_messages[key]

    async def run_until_all_text_messages_delivered(self, websocket_mock, timeout: int = 1):
        self.verify_async_init()
        key: uuid.UUID = get_stable_key(websocket_mock)
        all_delivered = self._all_incoming_websocket_text_delivered_event[key]
        await asyncio.wait_for(all_delivered.wait(), timeout)

    async def run_until_all_json_messages_delivered(self, websocket_mock, timeout: int = 1):
        self.verify_async_init()
        key: uuid.UUID = get_stable_key(websocket_mock)
        all_delivered = self._all_incoming_websocket_json_delivered_event[key]
        await asyncio.wait_for(all_delivered.wait(), timeout)

    async def run_until_all_aiohttp_messages_delivered(self, websocket_mock, timeout: int = 1):
        self.verify_async_init()
        key: uuid.UUID = get_stable_key(websocket_mock)
        all_delivered = self._all_incoming_websocket_aiohttp_delivered_event[key]
        await asyncio.wait_for(all_delivered.wait(), timeout)



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.bybit.com"
HEALTH_CHECK_ENDPOINT = "/v5/market/time"
CANDLES_ENDPOINT = "/v5/market/kline"

WSS_URL = "wss://stream.bybit.com/v5/public/spot"

INTERVALS = bidict({
    "1m": 1,
    "3m": 3,
    "5m": 5,
    "15m": 15,
    "30m": 30,
    "1h": 60,
    "2h": 120,
    "4h": 240,
    "6h": 360,
    "12h": 720,
    "1d": "D",
    "1w": "W",
    "1M": "M"
})

MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 1000

RATE_LIMITS = [
    RateLimit(CANDLES_ENDPOINT, limit=20000, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)]),
    RateLimit(HEALTH_CHECK_ENDPOINT, limit=20000, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)])]



from typing import Any, Callable, Dict, Optional

import hummingbot.connector.exchange.hashkey.hashkey_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.connector.utils import TimeSynchronizerRESTPreProcessor
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


def rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided public REST endpoint
    :param path_url: a public REST endpoint
    :param domain: the Hashkey domain to connect to ("mainnet" or "testnet"). The default value is "mainnet"
    :return: the full URL to the endpoint
    """
    return CONSTANTS.REST_URLS[domain] + path_url


def build_api_factory(
        throttler: Optional[AsyncThrottler] = None,
        time_synchronizer: Optional[TimeSynchronizer] = None,
        domain: str = CONSTANTS.DEFAULT_DOMAIN,
        time_provider: Optional[Callable] = None,
        auth: Optional[AuthBase] = None, ) -> WebAssistantsFactory:
    time_synchronizer = time_synchronizer or TimeSynchronizer()
    time_provider = time_provider or (lambda: get_current_server_time(
        throttler=throttler,
        domain=domain,
    ))
    throttler = throttler or create_throttler()
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        auth=auth,
        rest_pre_processors=[
            TimeSynchronizerRESTPreProcessor(synchronizer=time_synchronizer, time_provider=time_provider),
        ])
    return api_factory


def build_api_factory_without_time_synchronizer_pre_processor(throttler: AsyncThrottler) -> WebAssistantsFactory:
    api_factory = WebAssistantsFactory(throttler=throttler)
    return api_factory


def create_throttler() -> AsyncThrottler:
    return AsyncThrottler(CONSTANTS.RATE_LIMITS)


async def api_request(path: str,
                      api_factory: Optional[WebAssistantsFactory] = None,
                      throttler: Optional[AsyncThrottler] = None,
                      time_synchronizer: Optional[TimeSynchronizer] = None,
                      domain: str = CONSTANTS.DEFAULT_DOMAIN,
                      params: Optional[Dict[str, Any]] = None,
                      data: Optional[Dict[str, Any]] = None,
                      method: RESTMethod = RESTMethod.GET,
                      is_auth_required: bool = False,
                      return_err: bool = False,
                      limit_id: Optional[str] = None,
                      timeout: Optional[float] = None,
                      headers: Dict[str, Any] = {}):
    throttler = throttler or create_throttler()
    time_synchronizer = time_synchronizer or TimeSynchronizer()

    # If api_factory is not provided a default one is created
    # The default instance has no authentication capabilities and all authenticated requests will fail
    api_factory = api_factory or build_api_factory(
        throttler=throttler,
        time_synchronizer=time_synchronizer,
        domain=domain,
    )
    rest_assistant = await api_factory.get_rest_assistant()

    local_headers = {
        "Content-Type": "application/x-www-form-urlencoded"}
    local_headers.update(headers)
    url = rest_url(path, domain=domain)

    request = RESTRequest(
        method=method,
        url=url,
        params=params,
        data=data,
        headers=local_headers,
        is_auth_required=is_auth_required,
        throttler_limit_id=limit_id if limit_id else path
    )

    async with throttler.execute_task(limit_id=limit_id if limit_id else path):
        response = await rest_assistant.call(request=request, timeout=timeout)
        if response.status != 200:
            if return_err:
                error_response = await response.json()
                return error_response
            else:
                error_response = await response.text()
                if error_response is not None and "ret_code" in error_response and "ret_msg" in error_response:
                    raise IOError(f"The request to Hashkey failed. Error: {error_response}. Request: {request}")
                else:
                    raise IOError(f"Error executing request {method.name} {path}. "
                                  f"HTTP status is {response.status}. "
                                  f"Error: {error_response}")

        return await response.json()


async def get_current_server_time(
        throttler: Optional[AsyncThrottler] = None,
        domain: str = CONSTANTS.DEFAULT_DOMAIN,
) -> float:
    throttler = throttler or create_throttler()
    api_factory = build_api_factory_without_time_synchronizer_pre_processor(throttler=throttler)
    response = await api_request(
        path=CONSTANTS.SERVER_TIME_PATH_URL,
        api_factory=api_factory,
        throttler=throttler,
        domain=domain,
        method=RESTMethod.GET)
    server_time = response["serverTime"]

    return server_time



from decimal import Decimal
from typing import Any, Dict

from pydantic import ConfigDict, SecretStr
from pydantic.v1 import Field

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap, ClientFieldData
from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.0002"),
    taker_percent_fee_decimal=Decimal("0.0004"),
    buy_percent_fee_deducted_from_returns=True
)

CENTRALIZED = True

EXAMPLE_PAIR = "BTC-USDT"


def is_linear_perpetual(trading_pair: str) -> bool:
    """
    Returns True if trading_pair is in USDT(Linear) Perpetual
    """
    _, quote_asset = split_hb_trading_pair(trading_pair)
    return quote_asset in ["USDT", "USDC"]


def get_next_funding_timestamp(current_timestamp: float) -> float:
    # On Okx Perpetuals, funding occurs every 8 hours at 00:00UTC, 08:00UTC and 16:00UTC.
    # Reference: https://help.okx.com/hc/en-us/articles/360039261134-Funding-fee-calculation
    int_ts = int(current_timestamp)
    eight_hours = 8 * 60 * 60
    mod = int_ts % eight_hours
    return float(int_ts - mod + eight_hours)


def is_exchange_information_valid(rule: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information

    :param exchange_info: the exchange information for a trading pair

    :return: True if the trading pair is enabled, False otherwise
    """
    if "status" in rule and rule["status"] == "TRADING":
        valid = True
    else:
        valid = False
    return valid


class HashkeyPerpetualConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="hashkey_perpetual", client_data=None)
    hashkey_perpetual_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Hashkey Perpetual API key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    hashkey_perpetual_secret_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Hashkey Perpetual API secret",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )


KEYS = HashkeyPerpetualConfigMap.construct()

OTHER_DOMAINS = ["hashkey_perpetual_testnet"]
OTHER_DOMAINS_PARAMETER = {"hashkey_perpetual_testnet": "hashkey_perpetual_testnet"}
OTHER_DOMAINS_EXAMPLE_PAIR = {"hashkey_perpetual_testnet": "BTC-USDT"}
OTHER_DOMAINS_DEFAULT_FEES = {"hashkey_perpetual_testnet": [0.02, 0.04]}


class HashkeyPerpetualTestnetConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="hashkey_perpetual_testnet", client_data=None)
    hashkey_perpetual_testnet_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Hashkey Perpetual testnet API key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    hashkey_perpetual_testnet_secret_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Hashkey Perpetual testnet API secret",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    model_config = ConfigDict(title="hashkey_perpetual")


OTHER_DOMAINS_KEYS = {"hashkey_perpetual_testnet": HashkeyPerpetualTestnetConfigMap.construct()}
